# coding: utf-8
import json
import logging
import os
import platform
import re
import sys
import time
import traceback

try:
    from psutil import pid_exists
except ImportError:
    pid_exists = None

import requests
from requests.adapters import HTTPAdapter


logger = logging.getLogger('scrapydweb.utils.poll')  # __name__
_handler = logging.StreamHandler()
_formatter = logging.Formatter(fmt="[%(asctime)s] %(levelname)-8s in %(name)s: %(message)s")
_handler.setFormatter(_formatter)
logger.addHandler(_handler)

IN_WINDOWS = platform.system() == 'Windows'
JOB_PATTERN = re.compile(r"""
                            <tr>
                                <td>(?P<Project>.*?)</td>
                                <td>(?P<Spider>.*?)</td>
                                <td>(?P<Job>.*?)</td>
                                (?:<td>(?P<PID>.*?)</td>)?
                                (?:<td>(?P<Start>.*?)</td>)?
                                (?:<td>(?P<Runtime>.*?)</td>)?
                                (?:<td>(?P<Finish>.*?)</td>)?
                                (?:<td>(?P<Log>.*?)</td>)?
                                (?:<td>(?P<Items>.*?)</td>)?
                                [\w\W]*?  # Temp support for Scrapyd v1.3.0 (not released)
                            </tr>
                          """, re.X)
JOB_KEYS = ['project', 'spider', 'job', 'pid', 'start', 'runtime', 'finish', 'log', 'items']


class Poll(object):
    logger = logger

    def __init__(self, url_scrapydweb, username, password,
                 scrapyd_servers, scrapyd_servers_auths,
                 poll_round_interval, poll_request_interval,
                 main_pid, verbose, exit_timeout=0):
        self.url_scrapydweb = url_scrapydweb
        self.auth = (username, password) if username and password else None

        self.scrapyd_servers = scrapyd_servers
        self.scrapyd_servers_auths = scrapyd_servers_auths

        self.session = requests.Session()
        self.session.mount('http://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
        self.session.mount('https://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
        # if username and password:
            # self.session.auth = (username, password)
        self.timeout = 60

        self.poll_round_interval = poll_round_interval
        self.poll_request_interval = poll_request_interval

        self.ignore_finished_bool_list = [True] * len(self.scrapyd_servers)
        self.finished_jobs_dict = {}

        self.main_pid = main_pid
        self.poll_pid = os.getpid()

        if verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self.exit_timeout = exit_timeout

        self.init_time = time.time()
        self.url_stats = self.url_scrapydweb + '/{node}/log/{opt}/{project}/{spider}/{job}/?job_finished={job_finished}'

    def check_exit(self):
        exit_condition_1 = pid_exists is not None and not pid_exists(self.main_pid)
        exit_condition_2 = not IN_WINDOWS and not self.check_pid(self.main_pid)
        if exit_condition_1 or exit_condition_2:
            sys.exit("!!! Poll subprocess (pid: %s) exits "
                     "since main_pid %s not exists" % (self.poll_pid, self.main_pid))

    # https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
    @staticmethod
    def check_pid(pid):
        """ Check For the existence of a unix pid. """
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def fetch_jobs(self, node, url, auth):
        running_jobs = []
        finished_jobs_set = set()
        self.logger.debug("[node %s] fetch_jobs: %s", node, url)
        r = self.make_request(url, auth=auth, post=False)
        # Should not invoke update_finished_jobs() if fail to fetch jobs
        assert r is not None, "[node %s] fetch_jobs failed: %s" % (node, url)

        self.logger.debug("[node %s] fetch_jobs got (%s) %s bytes", node, r.status_code, len(r.content))
        # Temp support for Scrapyd v1.3.0 (not released)
        text = re.sub(r'<thead>.*?</thead>', '', r.text, flags=re.S)
        jobs = [dict(zip(JOB_KEYS, job)) for job in re.findall(JOB_PATTERN, text)]
        for job in jobs:
            job_tuple = (job['project'], job['spider'], job['job'])
            if job['pid']:
                running_jobs.append(job_tuple)
            elif job['finish']:
                finished_jobs_set.add(job_tuple)
        self.logger.debug("[node %s] got running_jobs: %s", node, len(running_jobs))
        self.logger.debug("[node %s] got finished_jobs_set: %s", node, len(finished_jobs_set))
        return running_jobs, finished_jobs_set

    def fetch_stats(self, node, job_tuple, finished_jobs):
        (project, spider, job) = job_tuple
        job_finished = 'True' if job_tuple in finished_jobs else ''
        kwargs = dict(
            node=node,
            opt='stats',
            project=project,
            spider=spider,
            job=job,
            job_finished=job_finished
        )
        # http://127.0.0.1:5000/log/stats/proxy/test/55f1f388a7ae11e8b9b114dda9e91c2f/
        url = self.url_stats.format(**kwargs)
        self.logger.debug("[node %s] fetch_stats: %s", node, url)
        # Make POST request to trigger alert, see log.py
        r = self.make_request(url, auth=self.auth, post=True)
        if r is None:
            self.logger.error("[node %s %s] fetch_stats failed: %s", node, self.scrapyd_servers[node-1], url)
            if job_finished:
                self.finished_jobs_dict[node].remove(job_tuple)
                self.logger.info("[node %s] retry in next round: %s", node, url)
        else:
            self.logger.debug("[node %s] fetch_stats got (%s) %s bytes from %s",
                              node, r.status_code, len(r.content), url)

    def main(self):
        while True:
            self.check_exit()
            start_time = time.time()
            try:
                self.run()
                end_time = time.time()
                self.logger.debug("Took %.1f seconds", (end_time - start_time))
                if 0 < self.exit_timeout < end_time - self.init_time:
                    self.logger.critical("GoodBye, exit_timeout: %s", self.exit_timeout)
                    break
                else:
                    self.logger.info("Sleeping for %ss", self.poll_round_interval)
                    time.sleep(self.poll_round_interval)
            except KeyboardInterrupt:
                self.logger.warning("Poll subprocess (pid: %s) cancelled by KeyboardInterrupt", self.poll_pid)
                sys.exit()
            except Exception:
                self.logger.error(traceback.format_exc())

    def make_request(self, url, auth, post=False):
        try:
            if post:
                r = self.session.post(url, auth=auth, timeout=self.timeout)
            else:
                r = self.session.get(url, auth=auth, timeout=self.timeout)
            r.encoding = 'utf-8'
            assert r.status_code == 200, "got status_code %s" % r.status_code
        except Exception as err:
            self.logger.error("make_request failed: %s\n%s", url, err)
            return None
        else:
            return r

    def run(self):
        for node, (scrapyd_server, auth) in enumerate(zip(self.scrapyd_servers, self.scrapyd_servers_auths), 1):
            # Update Jobs history
            # url_jobs = self.url_scrapydweb + '/%s/jobs/' % node
            # self.make_request(url_jobs, auth=self.auth, post=True)

            url_jobs = 'http://%s/jobs' % scrapyd_server
            # json.loads(json.dumps({'auth':(1,2)})) => {'auth': [1, 2]}
            auth = tuple(auth) if auth else None  # TypeError: 'list' object is not callable
            try:
                running_jobs, finished_jobs_set = self.fetch_jobs(node, url_jobs, auth)
                finished_jobs = self.update_finished_jobs(node, finished_jobs_set)
                for job_tuple in running_jobs + finished_jobs:
                    self.fetch_stats(node, job_tuple, finished_jobs)
                    self.logger.debug("Sleeping for %ss", self.poll_request_interval)
                    time.sleep(self.poll_request_interval)
            except KeyboardInterrupt:
                raise
            except AssertionError as err:
                self.logger.error(err)
            except Exception:
                self.logger.error(traceback.format_exc())

    def update_finished_jobs(self, node, finished_jobs_set):
        finished_jobs_set_previous = self.finished_jobs_dict.setdefault(node, set())
        self.logger.debug("[node %s] previous finished_jobs_set: %s", node, len(finished_jobs_set_previous))
        # set([2,3]).difference(set([1,2])) => {3}
        finished_jobs_set_new_added = finished_jobs_set.difference(finished_jobs_set_previous)
        self.finished_jobs_dict[node] = finished_jobs_set
        self.logger.debug("[node %s] now finished_jobs_set: %s", node, len(self.finished_jobs_dict[node]))
        if finished_jobs_set_new_added:
            self.logger.info("[node %s] new added finished_jobs_set: %s", node, finished_jobs_set_new_added)
        else:
            self.logger.debug("[node %s] new added finished_jobs_set: %s", node, finished_jobs_set_new_added)

        finished_jobs = []
        ignore = self.ignore_finished_bool_list[node-1]
        for job_tuple in finished_jobs_set_new_added:
            if ignore:
                self.logger.debug("[node %s] ignore finished job: %s", node, job_tuple)
            else:
                finished_jobs.append(job_tuple)
        if ignore:
            self.ignore_finished_bool_list[node-1] = False
            self.logger.debug("[node %s] new added finished_jobs after filter: %s", node, len(finished_jobs))
        return finished_jobs


def main(args):
    keys = ('url_scrapydweb', 'username', 'password',
            'scrapyd_servers', 'scrapyd_servers_auths',
            'poll_round_interval', 'poll_request_interval',
            'main_pid', 'verbose', 'exit_timeout')
    kwargs = dict(zip(keys, args))
    kwargs['scrapyd_servers'] = json.loads(kwargs['scrapyd_servers'])
    kwargs['scrapyd_servers_auths'] = json.loads(kwargs['scrapyd_servers_auths'])
    kwargs['poll_round_interval'] = int(kwargs['poll_round_interval'])
    kwargs['poll_request_interval'] = int(kwargs['poll_request_interval'])
    kwargs['main_pid'] = int(kwargs['main_pid'])
    kwargs['verbose'] = kwargs['verbose'] == 'True'
    kwargs['exit_timeout'] = int(kwargs.setdefault('exit_timeout', 0))  # For test only

    poll = Poll(**kwargs)
    poll.main()
    return poll.ignore_finished_bool_list  # For test only


if __name__ == '__main__':
    main(sys.argv[1:])



# coding: utf-8
from datetime import datetime
from io import BytesIO
import json
import os
import re
import time

from flask import url_for

from scrapydweb.utils.poll import main as poll_py_main
from tests.utils import cst, req, sleep, upload_file_deploy


def test_log_utf8_stats(app, client):
    # In ScrapydWeb_demo.egg: CONCURRENT_REQUESTS=1, DOWNLOAD_DELAY=10
    upload_file_deploy(app, client, filename='ScrapydWeb_demo.egg',
                       project=cst.PROJECT, redirect_project=cst.PROJECT)

    with app.test_request_context():
        kws = dict(node=1, opt='start', project=cst.PROJECT, version_spider_job=cst.SPIDER)
        __, js = req(app, client, view='api', kws=kws)
        jobid = js['jobid']
        sleep()

        # the Stats page
        req(app, client, view='log', kws=dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=jobid),
            ins='Log analysis')
        # the Log page
        req(app, client, view='log', kws=dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=jobid),
            ins='log - ScrapydWeb')

        # For testing request_scrapy_log() of LogView in log.py
        app.config['LOCAL_SCRAPYD_LOGS_DIR'] = 'non-exist-dir'
        req(app, client, view='log', kws=dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=jobid),
            ins='log - ScrapydWeb')

        # the Jobs page GET
        url_stop = url_for('api', node=1, opt='stop', project=cst.PROJECT, version_spider_job=jobid)
        url_jobs_classic = url_for('jobs', node=1, style='classic')
        url_jobs_database = url_for('jobs', node=1, style='database')
        req(app, client, view='jobs', kws=dict(node=1, style='classic'),
            ins=[url_stop, url_jobs_database, 'class="table wrap"'], nos="Vue.extend(Main)")
        req(app, client, view='jobs', kws=dict(node=1, style='database'),
            ins=[url_stop, url_jobs_classic, "Vue.extend(Main)"], nos='class="table wrap"')

        # ?raise_exception=True
        req(app, client, view='jobs', kws=dict(node=1, style='database'),
            ins=[url_stop, url_jobs_classic, "Vue.extend(Main)"], nos='class="table wrap"')
        req(app, client, view='jobs', kws=dict(node=1),
            ins=[url_stop, url_jobs_classic, "Vue.extend(Main)"], nos='class="table wrap"')
        req(app, client, view='metadata', kws=dict(node=1), jskws=dict(jobs_style='database'))

        req(app, client, view='jobs', kws=dict(node=1, raise_exception='True'),
            ins=[url_stop, url_jobs_database, 'class="table wrap"'], nos="Vue.extend(Main)")
        req(app, client, view='metadata', kws=dict(node=1), jskws=dict(jobs_style='classic'))
        req(app, client, view='jobs', kws=dict(node=1),
            ins=[url_stop, url_jobs_database, 'class="table wrap"'], nos="Vue.extend(Main)")

        # jobs POST data={}
        jobs_key = '%s/%s/%s' % (cst.PROJECT, cst.SPIDER, jobid)  # type unicode in Python 2
        print('######')
        print(repr(jobs_key))
        print(type(jobs_key))
        __, js = req(app, client, view='jobs', kws=dict(node=1), data={}, jskeys=jobs_key)
        jobs_id = js[jobs_key]['id']
        jobs_start = js[jobs_key]['start']
        assert js[jobs_key]['deleted'] == '0'

        # JobsXhrView delete running job
        req(app, client, view='jobs.xhr', kws=dict(node=1, action='delete', id=jobs_id),
            jskws=dict(status=cst.OK))
        # Recover deleted running job
        req(app, client, view='jobs', kws=dict(node=1, style='database'),
            ins=['Recover deleted job:', url_stop, 'id: %s,' % jobs_id, jobs_start])

        # forcestop
        client.get(url_for('api', node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=jobid))
        sleep()

        # /1/schedule/ScrapydWeb_demo/default:%20the%20latest%20version/test/   NOT unique
        url_start = url_for('schedule', node=1, project=cst.PROJECT,
                            version=cst.DEFAULT_LATEST_VERSION, spider=cst.SPIDER)
        req(app, client, view='jobs', kws=dict(node=1, style='classic'), ins=url_start)
        req(app, client, view='jobs', kws=dict(node=1, style='database'), ins=url_start)

        # JobsXhrView delete finished
        req(app, client, view='jobs.xhr', kws=dict(node=1, action='delete', id=jobs_id),
            jskws=dict(status=cst.OK))
        # JobsView: query_jobs(): self.jobs = self.Job.query.filter_by(deleted=NOT_DELETED)
        # POST data={}
        req(app, client, view='jobs', kws=dict(node=1), data={},
            nos=['id: %s,' % jobs_id, jobs_start])
        req(app, client, view='jobs', kws=dict(node=1, style='database'),
            nos=['id: %s,' % jobs_id, jobs_start])
        req(app, client, view='jobs', kws=dict(node=1, style='classic'), ins=jobs_start[5:])
        # delete id not exist
        req(app, client, view='jobs.xhr', kws=dict(node=1, action='delete', id=cst.BIGINT),
            jskws=dict(status=cst.ERROR))


def test_log_not_exist(app, client):
    # the Stats page
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.FAKE_JOBID)
    ins = ['fail - ScrapydWeb', 'status_code: 404']
    req(app, client, view='log', kws=kws, ins=ins)
    # the Log page
    kws['opt'] = 'utf8'
    ins = ['fail - ScrapydWeb', 'status_code: 404']
    req(app, client, view='log', kws=kws, ins=ins)


def test_inside_the_logs_page(app, client):
    with app.test_request_context():
        for project, spider in [(cst.PROJECT, None), (cst.PROJECT, cst.SPIDER)]:
            title = 'Directory listing for /logs/%s/%s' % (project, spider or '')

            text, __ = req(app, client, view='logs', kws=dict(node=1, project=project, spider=spider), ins=title)

            if spider:
                # http://127.0.0.1:6800/logs/ScrapydWeb_demo/test/ScrapydWeb_demo.log
                scrapyd_server = app.config['SCRAPYD_SERVERS'][0]
                assert 'http://%s/logs/%s/%s/%s' % (scrapyd_server, cst.PROJECT, cst.SPIDER, cst.DEMO_LOG) in text
                assert 'http://%s/logs/%s/%s/%s' % (scrapyd_server, cst.PROJECT, cst.SPIDER, cst.DEMO_JSON) in text

                # http://127.0.0.1:5000/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo.json/?with_ext=True
                url_stats_of_json = url_for('log', node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER,
                                            job=cst.DEMO_JSON, with_ext='True')
                assert url_stats_of_json in text
                # http://127.0.0.1:5000/1/log/utf8/ScrapydWeb_demo/test/ScrapydWeb_demo.json/?with_ext=True
                url_utf8_of_json = url_for('log', node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER,
                                           job=cst.DEMO_JSON, with_ext='True')
                assert url_utf8_of_json not in text

                url_run_spider = url_for('schedule', node=1, project=cst.PROJECT,
                                         version=cst.DEFAULT_LATEST_VERSION, spider=cst.SPIDER)
                assert url_run_spider in text


# Links of Stats and Log in the Logs page
# http://127.0.0.1:5000/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo.log/?with_ext=True
# http://127.0.0.1:5000/1/log/utf8/ScrapydWeb_demo/test/ScrapydWeb_demo.log/?with_ext=True
# Source: http://127.0.0.1:6800/logs/ScrapydWeb_demo/test/ScrapydWeb_demo.log
def test_demo_log_with_extension(app, client):
    with app.test_request_context():
        url_utf8 = url_for('log', node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER,
                           job=cst.DEMO_LOG, with_ext='True')
        url_stats = url_for('log', node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER,
                            job=cst.DEMO_LOG, with_ext='True')
        scrapyd_server = app.config['SCRAPYD_SERVERS'][0]
        url_demo_log_source = 'http://%s/logs/%s/%s/%s' % (scrapyd_server, cst.PROJECT, cst.SPIDER, cst.DEMO_LOG)
        # the Stats page
        kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_LOG, with_ext='True')
        ins = ['<tr><th>runtime</th><td>0:01:08</td></tr>', 'id="finish_reason">finished<',
               '<h4>Log</h4>', url_utf8, '<h4>Source</h4>', url_demo_log_source]
        req(app, client, view='log', kws=kws, ins=ins)
        # the Log page
        kws = dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_LOG, with_ext='True')
        ins = ['the Stats page &gt;&gt; View log &gt;&gt; Tail', 'PROJECT (ScrapydWeb_demo), SPIDER (test)',
               '<h4>Stats</h4>', url_stats, '<h4>Source</h4>', url_demo_log_source]
        req(app, client, view='log', kws=kws, ins=ins)

        # Stats link of json file in the Logs page
        # http://127.0.0.1:5000/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo.json/?with_ext=True
        url_utf8_ = url_for('log', node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER,
                            job=cst.DEMO_JSON, with_ext='True')
        url_demo_json_source = 'http://%s/logs/%s/%s/%s' % (scrapyd_server, cst.PROJECT, cst.SPIDER, cst.DEMO_JSON)
        kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JSON, with_ext='True')
        ins = ['<tr><th>runtime</th><td>0:01:08</td></tr>', 'id="finish_reason">finished<']
        req(app, client, view='log', kws=kws, ins=ins,
            nos=['<h4>Log</h4>', url_utf8_, '<h4>Source</h4>', url_demo_json_source])


# Location: http://127.0.0.1:5000/log/uploaded/ttt.txt
def test_parse_upload(app, client):
    req(app, client, view='parse.upload', kws=dict(node=1),
        data={'file': (BytesIO(b''), 'empty.log')},
        location='/parse/uploaded/')

    req(app, client, view='parse.upload', kws=dict(node=1),
        data={'file': (BytesIO(b'my file contents'), 'invalid.txt')},
        location='/parse/uploaded/')

    req(app, client, view='parse.upload', kws=dict(node=1),
        data={'file': (os.path.join(cst.ROOT_DIR, 'data/%s' % cst.DEMO_LOG), cst.DEMO_LOG)},
        location='/parse/uploaded/')


def test_parse_uploaded_empty_invalid_log(app, client):
    for filename in ['empty.log', 'invalid.txt']:
        req(app, client, view='parse.uploaded', kws=dict(node=1, filename=filename),
            ins=['<tr><th>job</th><td>%s</td></tr>' % filename.split('.')[0],
                 '<tr><th>first_log_time</th><td>N/A</td></tr>'])


def test_parse_uploaded_demo_log(app, client):
    # Time string extracted from logfile doesnot contains timezone info, so avoid using hard coding timestamp.
    def string_to_timestamp(string):
        datetime_obj = datetime.strptime(string, '%Y-%m-%d %H:%M:%S')
        return int(time.mktime(datetime_obj.timetuple()))
    latest_crawl_timestamp = string_to_timestamp('2018-10-23 18:28:39')
    latest_scrape_timestamp = string_to_timestamp('2018-10-23 18:28:39')
    latest_log_timestamp = string_to_timestamp('2018-10-23 18:29:42')
    ins = [
        'PROJECT (demo), SPIDER (test)',
        '<tr><th>job</th><td>2018-10-23_182826</td></tr>',
        'Log analysis',
        '0:01:08',
        '"green">3<',
        '"green">2<',
        'id="finish_reason">finished<',
        'id="log_critical_count">5<',
        'id="log_error_count">5<',
        'id="log_warning_count">3<',
        'id="log_redirect_count">1<',
        'id="log_retry_count">2<',
        'id="log_ignore_count">1<',
        'var latest_crawl_timestamp = %s;' % latest_crawl_timestamp,  # 1540290519
        'var latest_scrape_timestamp = %s;' % latest_scrape_timestamp,  # 1540290519
        'var latest_log_timestamp = %s;' % latest_log_timestamp  # 1540290582
    ]
    text, __ = req(app, client, view='parse.uploaded', kws=dict(node=1, filename=cst.DEMO_LOG), ins=ins)
    # const LAST_UPDATE_TIMESTAMP = 1547708035;
    assert time.time() - int(re.search(r'LAST_UPDATE_TIMESTAMP = (\d+)', text).group(1)) < 3


def test_parse_source_demo_log(app, client):
    req(app, client, view='parse.source', kws=dict(filename=cst.DEMO_LOG),
        ins=['2018-10-23 18:28:34 [scrapy.utils.log] INFO: Scrapy 1.5.0 started (bot: demo)',
             '2018-10-23 18:29:42 [scrapy.core.engine] INFO: Spider closed (finished)'])


# scrapydweb/scrapydweb/utils/sub_process.py
def test_poll_py(app):
    _bind = app.config.get('SCRAPYDWEB_BIND', '0.0.0.0')
    _bind = '127.0.0.1' if _bind == '0.0.0.0' else _bind
    args = [
        'http://%s:%s' % (_bind, app.config.get('SCRAPYDWEB_PORT', 5000)),
        app.config.get('USERNAME', '') if app.config.get('ENABLE_AUTH', False) else '',
        app.config.get('PASSWORD', '') if app.config.get('ENABLE_AUTH', False) else '',
        json.dumps(app.config.get('SCRAPYD_SERVERS', ['127.0.0.1'])),
        json.dumps(app.config.get('SCRAPYD_SERVERS_AUTHS', [None])),
        '3',  # str(app.config.get('POLL_ROUND_INTERVAL', 300)),
        '1',  # str(app.config.get('POLL_REQUEST_INTERVAL', 10)),
        str(app.config['MAIN_PID']),
        str(app.config.get('VERBOSE', False)),
        '10'  # exit_timeout
    ]
    ignore_finished_bool_list = poll_py_main(args)
    assert ignore_finished_bool_list == [False, True]


def test_monitor_alert(app, client):
    # In ScrapydWeb_demo_no_delay.egg: unset CONCURRENT_REQUESTS, unset DOWNLOAD_DELAY
    upload_file_deploy(app, client, filename='ScrapydWeb_demo_no_delay.egg',
                       project=cst.PROJECT, redirect_project=cst.PROJECT)

    # with app.test_request_context():
    if not (app.config.get('ENABLE_MONITOR', False) and app.config.get('ENABLE_EMAIL_ALERT', False)):
        return

    def start_a_job():
        kws = dict(node=1, opt='start', project=cst.PROJECT, version_spider_job=cst.SPIDER)
        __, js = req(app, client, view='api', kws=kws)
        sleep()
        return js['jobid']

    def forcestop_a_job(job):
        req(app, client, view='api', kws=dict(node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=job))

    def post_for_poll(job, job_finished=''):
        kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=job, job_finished=job_finished)
        req(app, client, view='log', kws=kws, data={}, ins='Log analysis')

    # Simulate poll post 'Finished'
    app.config['ON_JOB_FINISHED'] = True
    jobid = start_a_job()
    post_for_poll(jobid, job_finished='True')
    forcestop_a_job(jobid)
    sleep()

    # Simulate poll post 'ForceStopped'
    app.config['ON_JOB_FINISHED'] = False
    app.config['LOG_CRITICAL_THRESHOLD'] = 1
    app.config['LOG_CRITICAL_TRIGGER_FORCESTOP'] = True
    jobid = start_a_job()
    post_for_poll(jobid)
    forcestop_a_job(jobid)
    sleep()

    # Simulate poll post 'Stopped'
    app.config['LOG_CRITICAL_THRESHOLD'] = 0
    app.config['LOG_REDIRECT_THRESHOLD'] = 1
    app.config['LOG_REDIRECT_TRIGGER_STOP'] = True
    jobid = start_a_job()
    post_for_poll(jobid)
    forcestop_a_job(jobid)
    sleep()

    # Simulate poll post 'Triggered'
    app.config['LOG_REDIRECT_THRESHOLD'] = 0
    app.config['LOG_IGNORE_THRESHOLD'] = 1
    jobid = start_a_job()
    post_for_poll(jobid)
    forcestop_a_job(jobid)
    sleep()

    # Simulate poll post 'Running'
    app.config['LOG_IGNORE_THRESHOLD'] = 0
    app.config['ON_JOB_RUNNING_INTERVAL'] = 5
    jobid = start_a_job()
    post_for_poll(jobid)  # Would not trigger email

    sleep()
    post_for_poll(jobid)  # Would trigger email

    app.config['ON_JOB_RUNNING_INTERVAL'] = 0
    sleep()
    post_for_poll(jobid)  # Would not trigger email
    forcestop_a_job(jobid)



# coding: utf-8
from collections import OrderedDict, defaultdict
import re

from flask import render_template
from logparser import SETTINGS_PY_PATH as LOGPARSER_SETTINGS_PY_PATH

from ...common import json_dumps
from ...vars import SCHEDULER_STATE_DICT
from ..baseview import BaseView


class SettingsView(BaseView):
    methods = ['GET']

    def __init__(self):
        super(SettingsView, self).__init__()

        self.template = 'scrapydweb/settings.html'
        self.kwargs = dict(node=self.node)

    def dispatch_request(self, **kwargs):
        self.update_kwargs()
        return render_template(self.template, **self.kwargs)

    @staticmethod
    def json_dumps(obj, sort_keys=False):
        string = json_dumps(obj, sort_keys=sort_keys)
        return string.replace(' true', ' True').replace(' false', ' False').replace(' null', ' None')

    @staticmethod
    def protect(string):
        if not isinstance(string, str):
            return string
        length = len(string)
        if length < 4:
            return '*' * length
        elif length < 12:
            return ''.join([string[i] if not i%2 else '*' for i in range(0, length)])
        else:
            return re.sub(r'^.{4}(.*?).{4}$', r'****\1****', string)

    @staticmethod
    def hide_account(string):
        return re.sub(r'//.+@', '//', string)

    def update_kwargs(self):
        # User settings
        self.kwargs['DEFAULT_SETTINGS_PY_PATH'] = self.handle_slash(self.DEFAULT_SETTINGS_PY_PATH)
        self.kwargs['SCRAPYDWEB_SETTINGS_PY_PATH'] = self.handle_slash(self.SCRAPYDWEB_SETTINGS_PY_PATH)
        self.kwargs['MAIN_PID'] = self.MAIN_PID
        self.kwargs['LOGPARSER_PID'] = self.LOGPARSER_PID
        self.kwargs['POLL_PID'] = self.POLL_PID

        # ScrapydWeb
        self.kwargs['scrapydweb_server'] = self.json_dumps(dict(
            SCRAPYDWEB_BIND=self.SCRAPYDWEB_BIND,
            SCRAPYDWEB_PORT=self.SCRAPYDWEB_PORT,
            URL_SCRAPYDWEB=self.URL_SCRAPYDWEB,
            ENABLE_AUTH=self.ENABLE_AUTH,
            USERNAME=self.protect(self.USERNAME),
            PASSWORD=self.protect(self.PASSWORD)
        ))
        self.kwargs['ENABLE_HTTPS'] = self.ENABLE_HTTPS
        self.kwargs['enable_https_details'] = self.json_dumps(dict(
            CERTIFICATE_FILEPATH=self.CERTIFICATE_FILEPATH,
            PRIVATEKEY_FILEPATH=self.PRIVATEKEY_FILEPATH
        ))

        # Scrapy
        self.kwargs['SCRAPY_PROJECTS_DIR'] = self.handle_slash(self.SCRAPY_PROJECTS_DIR) or "''"

        # Scrapyd
        servers = defaultdict(list)
        for group, server, auth in zip(self.SCRAPYD_SERVERS_GROUPS, self.SCRAPYD_SERVERS, self.SCRAPYD_SERVERS_AUTHS):
            _server = '%s:%s@%s' % (self.protect(auth[0]), self.protect(auth[1]), server) if auth else server
            servers[group].append(_server)

        self.kwargs['servers'] = self.json_dumps(servers)
        self.kwargs['LOCAL_SCRAPYD_SERVER'] = self.LOCAL_SCRAPYD_SERVER or "''"
        self.kwargs['LOCAL_SCRAPYD_LOGS_DIR'] = self.handle_slash(self.LOCAL_SCRAPYD_LOGS_DIR) or "''"
        self.kwargs['SCRAPYD_LOG_EXTENSIONS'] = self.SCRAPYD_LOG_EXTENSIONS

        # LogParser
        self.kwargs['ENABLE_LOGPARSER'] = self.ENABLE_LOGPARSER
        self.kwargs['logparser_version'] = self.LOGPARSER_VERSION
        self.kwargs['logparser_settings_py_path'] = self.handle_slash(LOGPARSER_SETTINGS_PY_PATH)
        self.kwargs['BACKUP_STATS_JSON_FILE'] = self.BACKUP_STATS_JSON_FILE

        # Timer Tasks
        self.kwargs['scheduler_state'] = SCHEDULER_STATE_DICT[self.scheduler.state]
        self.kwargs['JOBS_SNAPSHOT_INTERVAL'] = self.JOBS_SNAPSHOT_INTERVAL

        # Run Spider
        self.kwargs['run_spider_details'] = self.json_dumps(dict(
            SCHEDULE_EXPAND_SETTINGS_ARGUMENTS=self.SCHEDULE_EXPAND_SETTINGS_ARGUMENTS,
            SCHEDULE_CUSTOM_USER_AGENT=self.SCHEDULE_CUSTOM_USER_AGENT,
            SCHEDULE_USER_AGENT=self.SCHEDULE_USER_AGENT,
            SCHEDULE_ROBOTSTXT_OBEY=self.SCHEDULE_ROBOTSTXT_OBEY,
            SCHEDULE_COOKIES_ENABLED=self.SCHEDULE_COOKIES_ENABLED,
            SCHEDULE_CONCURRENT_REQUESTS=self.SCHEDULE_CONCURRENT_REQUESTS,
            SCHEDULE_DOWNLOAD_DELAY=self.SCHEDULE_DOWNLOAD_DELAY,
            SCHEDULE_ADDITIONAL=self.SCHEDULE_ADDITIONAL
        ))

        # Page Display
        self.kwargs['page_display_details'] = self.json_dumps(dict(
            SHOW_SCRAPYD_ITEMS=self.SHOW_SCRAPYD_ITEMS,
            SHOW_JOBS_JOB_COLUMN=self.SHOW_JOBS_JOB_COLUMN,
            JOBS_FINISHED_JOBS_LIMIT=self.JOBS_FINISHED_JOBS_LIMIT,
            JOBS_RELOAD_INTERVAL=self.JOBS_RELOAD_INTERVAL,
            DAEMONSTATUS_REFRESH_INTERVAL=self.DAEMONSTATUS_REFRESH_INTERVAL
        ))

        # Send text
        self.kwargs['slack_details'] = self.json_dumps(dict(
            SLACK_TOKEN=self.protect(self.SLACK_TOKEN),
            SLACK_CHANNEL=self.SLACK_CHANNEL
        ))
        self.kwargs['telegram_details'] = self.json_dumps(dict(
            TELEGRAM_TOKEN=self.protect(self.TELEGRAM_TOKEN),
            TELEGRAM_CHAT_ID=self.TELEGRAM_CHAT_ID
        ))
        self.kwargs['email_details'] = self.json_dumps(dict(
            EMAIL_SUBJECT=self.EMAIL_SUBJECT,
        ))
        self.kwargs['email_sender_recipients'] = self.json_dumps(dict(
            EMAIL_USERNAME=self.EMAIL_USERNAME,
            EMAIL_PASSWORD=self.protect(self.EMAIL_PASSWORD),
            EMAIL_SENDER=self.EMAIL_SENDER,
            EMAIL_RECIPIENTS=self.EMAIL_RECIPIENTS
        ))
        self.kwargs['email_smtp_settings'] = self.json_dumps(dict(
            SMTP_SERVER=self.SMTP_SERVER,
            SMTP_PORT=self.SMTP_PORT,
            SMTP_OVER_SSL=self.SMTP_OVER_SSL,
            SMTP_CONNECTION_TIMEOUT=self.SMTP_CONNECTION_TIMEOUT,
        ))


        # Monitor & Alert
        self.kwargs['ENABLE_MONITOR'] = self.ENABLE_MONITOR
        self.kwargs['poll_interval'] = self.json_dumps(dict(
            POLL_ROUND_INTERVAL=self.POLL_ROUND_INTERVAL,
            POLL_REQUEST_INTERVAL=self.POLL_REQUEST_INTERVAL,
        ))
        self.kwargs['alert_switcher'] = self.json_dumps(dict(
            ENABLE_SLACK_ALERT=self.ENABLE_SLACK_ALERT,
            ENABLE_TELEGRAM_ALERT=self.ENABLE_TELEGRAM_ALERT,
            ENABLE_EMAIL_ALERT=self.ENABLE_EMAIL_ALERT,
        ))
        self.kwargs['alert_working_time'] = self.json_dumps([
            dict(
                ALERT_WORKING_DAYS="%s" % sorted(self.ALERT_WORKING_DAYS),  # stringify making it displayed in a line
                remark="Monday is 1 and Sunday is 7"
            ),
            dict(
                ALERT_WORKING_HOURS="%s" % sorted(self.ALERT_WORKING_HOURS),
                remark="From 0 to 23"
            )
        ])
        # alert triggers
        d = OrderedDict()
        d['ON_JOB_RUNNING_INTERVAL'] = self.ON_JOB_RUNNING_INTERVAL
        d['ON_JOB_FINISHED'] = self.ON_JOB_FINISHED

        for key in self.ALERT_TRIGGER_KEYS:
            keys = ['LOG_%s_THRESHOLD' % key, 'LOG_%s_TRIGGER_STOP' % key, 'LOG_%s_TRIGGER_FORCESTOP' % key]
            d[key] = {k: getattr(self, k) for k in keys}
        value = self.json_dumps(d)
        value = re.sub(r'True', "<b style='color: red'>True</b>", value)
        value = re.sub(r'(\s[1-9]\d*)', r"<b style='color: red'>\1</b>", value)
        self.kwargs['alert_triggers'] = value

        # System
        self.kwargs['DEBUG'] = self.DEBUG
        self.kwargs['VERBOSE'] = self.VERBOSE
        self.kwargs['DATA_PATH'] = self.DATA_PATH
        self.kwargs['database_details'] = self.json_dumps(dict(
            APSCHEDULER_DATABASE_URI=self.hide_account(self.APSCHEDULER_DATABASE_URI),
            SQLALCHEMY_DATABASE_URI=self.hide_account(self.SQLALCHEMY_DATABASE_URI),
            SQLALCHEMY_BINDS_METADATA=self.hide_account(self.SQLALCHEMY_BINDS['metadata']),
            SQLALCHEMY_BINDS_JOBS=self.hide_account(self.SQLALCHEMY_BINDS['jobs'])
        ))



# coding: utf-8
import os
import re

from flask import render_template, url_for

from ...vars import DIRECTORY_KEYS, DIRECTORY_PATTERN, HREF_NAME_PATTERN
from ..baseview import BaseView


class ItemsView(BaseView):
    methods = ['GET']

    def __init__(self):
        super(ItemsView, self).__init__()

        self.project = self.view_args['project']
        self.spider = self.view_args['spider']

        self.url = 'http://{}/items/{}{}'.format(self.SCRAPYD_SERVER,
                                                 '%s/' % self.project if self.project else '',
                                                 '%s/' % self.spider if self.spider else '')
        if self.SCRAPYD_SERVER_PUBLIC_URL:
            self.public_url = re.sub(r'^http.*?/items/', self.SCRAPYD_SERVER_PUBLIC_URL + '/items/', self.url)
        else:
            self.public_url = ''
        self.template = 'scrapydweb/logs_items.html'
        self.text = ''

    def dispatch_request(self, **kwargs):
        status_code, self.text = self.make_request(self.url, auth=self.AUTH, as_json=False)
        if status_code != 200 or not re.search(r'Directory listing for /items/', self.text):
            kwargs = dict(
                node=self.node,
                url=self.url,
                status_code=status_code,
                text=self.text,
                tip="Click the above link to make sure your Scrapyd server is accessable. "
            )
            return render_template(self.template_fail, **kwargs)

        return self.generate_response()

    def generate_response(self):
        rows = [dict(zip(DIRECTORY_KEYS, row)) for row in re.findall(DIRECTORY_PATTERN, self.text)]
        for row in rows:
            # <a href="demo/">demo/</a>     dir
            # <a href="test/">test/</a>     dir
            # <a href="a.jl">a.jl</a>       file
            row['href'], row['filename'] = re.search(HREF_NAME_PATTERN, row['filename']).groups()
            if not row['href'].endswith('/'):  # It's a file but not a directory
                row['href'] = (self.public_url or self.url) + row['href']

            if self.project and self.spider:
                if row['filename'].endswith('.tar.gz'):
                    filename_without_ext = row['filename'][:-len('.tar.gz')]
                else:
                    filename_without_ext = os.path.splitext(row['filename'])[0]  # '1.1.jl' => ('1.1', '.jl')
                row['url_stats'] = url_for('log', node=self.node, opt='stats', project=self.project,
                                           spider=self.spider, job=filename_without_ext)
                row['url_utf8'] = url_for('log', node=self.node, opt='utf8', project=self.project,
                                          spider=self.spider, job=filename_without_ext)
                row['url_clusterreports'] = url_for('clusterreports', node=self.node, project=self.project,
                                                    spider=self.spider, job=self.get_job_without_ext(row['filename']))
        if self.project and self.spider:
            url_schedule = url_for('schedule', node=self.node, project=self.project,
                                   version=self.DEFAULT_LATEST_VERSION, spider=self.spider)
            url_multinode_run = url_for('servers', node=self.node, opt='schedule', project=self.project,
                                        version_job=self.DEFAULT_LATEST_VERSION, spider=self.spider)
        else:
            url_schedule = url_multinode_run = ''
        kwargs = dict(
            node=self.node,
            title='items',
            project=self.project,
            spider=self.spider,
            url=self.url,
            url_schedule=url_schedule,
            url_multinode_run=url_multinode_run,
            rows=rows
        )
        return render_template(self.template, **kwargs)



# coding: utf-8
import glob
import importlib
import io
import os
import re
import sys

from apscheduler.schedulers.base import STATE_PAUSED, STATE_RUNNING, STATE_STOPPED

from .default_settings import DATA_PATH as default_data_path
from .default_settings import DATABASE_URL as default_database_url
from .utils.setup_database import setup_database


PYTHON_VERSION = '.'.join([str(n) for n in sys.version_info[:3]])
PY2 = sys.version_info.major < 3
SCRAPYDWEB_SETTINGS_PY = 'scrapydweb_settings_v10.py'
sys.path.append(os.getcwd())
try:
    custom_settings_module = importlib.import_module(os.path.splitext(SCRAPYDWEB_SETTINGS_PY)[0])
except ImportError:
    custom_data_path = ''
    custom_database_url = ''
else:
    custom_data_path = getattr(custom_settings_module, 'DATA_PATH', '')
    custom_data_path = custom_data_path if isinstance(custom_data_path, str) else ''
    custom_database_url = getattr(custom_settings_module, 'DATABASE_URL', '')
    custom_database_url = custom_database_url if isinstance(custom_database_url, str) else ''

# For data storage
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

DATA_PATH = default_data_path or custom_data_path
if DATA_PATH:
    DATA_PATH = os.path.abspath(DATA_PATH)
else:
    DATA_PATH = os.path.join(ROOT_DIR, 'data')

DATABASE_PATH = os.path.join(DATA_PATH, 'database')
DEMO_PROJECTS_PATH = os.path.join(DATA_PATH, 'demo_projects')
DEPLOY_PATH = os.path.join(DATA_PATH, 'deploy')
HISTORY_LOG = os.path.join(DATA_PATH, 'history_log')
PARSE_PATH = os.path.join(DATA_PATH, 'parse')
SCHEDULE_PATH = os.path.join(DATA_PATH, 'schedule')
STATS_PATH = os.path.join(DATA_PATH, 'stats')

for path in [DATA_PATH, DATABASE_PATH, DEMO_PROJECTS_PATH, DEPLOY_PATH,
             HISTORY_LOG, PARSE_PATH, SCHEDULE_PATH, STATS_PATH]:
    if not os.path.isdir(path):
        os.mkdir(path)
    elif path in [PARSE_PATH, DEPLOY_PATH, SCHEDULE_PATH]:
        for file in glob.glob(os.path.join(path, '*.*')):
            if not os.path.split(file)[-1] in ['ScrapydWeb_demo.log']:
                os.remove(file)

RUN_SPIDER_HISTORY_LOG = os.path.join(HISTORY_LOG, 'run_spider_history.log')
TIMER_TASKS_HISTORY_LOG = os.path.join(HISTORY_LOG, 'timer_tasks_history.log')

# For database
DATABASE_URL = custom_database_url or default_database_url or 'sqlite:///' + DATABASE_PATH
results = setup_database(DATABASE_URL, DATABASE_PATH)
APSCHEDULER_DATABASE_URI, SQLALCHEMY_DATABASE_URI, SQLALCHEMY_BINDS, DATABASE_PATH = results

# For check_app_config() and BaseView
ALLOWED_SCRAPYD_LOG_EXTENSIONS = ['.log', '.log.gz', '.txt', '.gz', '']
ALERT_TRIGGER_KEYS = ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']

# Error: Project names must begin with a letter and contain only letters, numbers and underscores
STRICT_NAME_PATTERN = re.compile(r'[^0-9A-Za-z_]')
LEGAL_NAME_PATTERN = re.compile(r'[^0-9A-Za-z_-]')

# For ScheduleView
SCHEDULE_ADDITIONAL = "-d setting=CLOSESPIDER_TIMEOUT=60\r\n-d setting=CLOSESPIDER_PAGECOUNT=10\r\n-d arg1=val1"
UA_DICT = {
    'custom': "Mozilla/5.0",
    'Chrome': ("Mozilla/5.0 (Windows NT 10.0; WOW64) "
               "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.109 Safari/537.36"),
    'iPhone': ("Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) "
               "AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1"),
    'iPad': ("Mozilla/5.0 (iPad; CPU OS 12_1_4 like Mac OS X) "
             "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1"),
    'Android': ("Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2 XL Build/OPD1.170816.004) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.109 Mobile Safari/537.36")
}

# For LogsView and ItemsView
DIRECTORY_PATTERN = re.compile(r"""
                                    <tr\sclass="(?P<odd_even>odd|even)">\n
                                        \s+<td>(?P<filename>.*?)</td>\n
                                        \s+<td>(?P<size>.*?)</td>\n
                                        \s+<td>(?P<content_type>.*?)</td>\n
                                        \s+<td>(?P<content_encoding>.*?)</td>\n
                                    </tr>
                                """, re.X)
DIRECTORY_KEYS = ['odd_even', 'filename', 'size', 'content_type', 'content_encoding']
HREF_NAME_PATTERN = re.compile(r'href="(.+?)">(.+?)<')

# For JobsView
jobs_table_map = {}

# For Timer Tasks
# STATE_STOPPED = 0, STATE_RUNNING = 1, STATE_PAUSED = 2
SCHEDULER_STATE_DICT = {
    STATE_STOPPED: 'STATE_STOPPED',
    STATE_RUNNING: 'STATE_RUNNING',
    STATE_PAUSED: 'STATE_PAUSED',
}


def setup_logfile(delete=False):
    if delete:
        for logfile in [RUN_SPIDER_HISTORY_LOG, TIMER_TASKS_HISTORY_LOG]:
            if os.path.exists(logfile):
                os.remove(logfile)

    if not os.path.exists(RUN_SPIDER_HISTORY_LOG):
        with io.open(RUN_SPIDER_HISTORY_LOG, 'w', encoding='utf-8') as f:
            f.write(u'%s\n%s' % ('#' * 50, RUN_SPIDER_HISTORY_LOG))

    if not os.path.exists(TIMER_TASKS_HISTORY_LOG):
        with io.open(TIMER_TASKS_HISTORY_LOG, 'w', encoding='utf-8') as f:
            f.write(u'%s\n%s\n\n' % (TIMER_TASKS_HISTORY_LOG, '#' * 50))


setup_logfile(delete=False)



# coding: utf-8
import time
from flask import url_for

from tests.utils import cst, req_single_scrapyd, sleep, upload_file_deploy


metadata = dict(value=time.ctime())
FILENAME = '%s_%s_%s.pickle' % (cst.PROJECT, cst.VERSION, cst.SPIDER)

# http://flask.pocoo.org/docs/1.0/tutorial/tests/#id11
# def test_author_required(app, client, auth):
# http://flask.pocoo.org/docs/1.0/testing/#other-testing-tricks


# START button in Jobs | Run Spider button in Logs
# this.form.selectedProject = 'ScrapydWeb_demo';
# this.form.selectedVersion = 'default: the latest version';
# this.loadSpiders();
# this.form.selectedSpider = 'test';
def test_schedule_with_default_version(app, client):
    ins = [
        "selectedProject = '%s'" % cst.PROJECT,
        "selectedVersion = 'default: the latest version'",
        "this.loadSpiders();",
        "selectedSpider = '%s'" % cst.SPIDER
    ]
    kws = dict(node=1, project=cst.PROJECT, version=cst.DEFAULT_LATEST_VERSION, spider=cst.SPIDER)
    req_single_scrapyd(app, client, view='schedule', kws=kws, ins=ins)


# {
# "project": "demo",
# "_version": "2018-01-01T01_01_01",
# "spider": "test"
# "jobid": "2018-12-01T15_32_01",
# "USER_AGENT": "Chrome",
# "COOKIES_ENABLED": "False",
# "ROBOTSTXT_OBEY": "False",
# "CONCURRENT_REQUESTS": "1",
# "DOWNLOAD_DELAY": "2",
# "additional": "-d setting=CLOSESPIDER_TIMEOUT=60 \r\n-d setting=CLOSESPIDER_PAGECOUNT=10 \r\n-d arg1=val1",
# }

# {
# "project": "ScrapydWeb_demo",
# "_version": "default: the latest version",
# "spider": "test"
# }
def test_check(app, client):
    # ScrapydWeb_demo.egg: custom_settings = {}, also log settings & arguments
    upload_file_deploy(app, client, filename='ScrapydWeb_demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT)
    data = dict(
        project=cst.PROJECT,
        _version=cst.VERSION,
        spider=cst.SPIDER,
        jobid=cst.JOBID,
        USER_AGENT='Chrome',
        ROBOTSTXT_OBEY='False',
        COOKIES_ENABLED='False',
        CONCURRENT_REQUESTS='1',
        DOWNLOAD_DELAY='2',
        additional=("-d setting=CLOSESPIDER_TIMEOUT=60 \r\n-d setting=CLOSESPIDER_PAGECOUNT=10 \r\n"
                    "-d arg1=%s") % metadata['value']
    )

    data_ = dict(
        project=cst.PROJECT,
        _version=cst.DEFAULT_LATEST_VERSION,
        spider=cst.SPIDER,
        additional="-d setting=CLOSESPIDER_TIMEOUT=60 -d arg1"
    )
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=1), data=data,
                       jskws=dict(filename=FILENAME))
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=1), data=data_,
                       jskws=dict(filename='%s_%s_%s.pickle' % (cst.PROJECT, 'default-the-latest-version', cst.SPIDER)))


# {
# "1": "on",
# "checked_amount": "1",
# "filename": "demo_2018-10-27T16_17_43_test.pickle"
# }
# <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n<title>Redirecting...</title>\n
# <h1>Redirecting...</h1>\n<p>You should be redirected automatically to target URL:
# <a href="/1/jobs/">/1/jobs/</a>.  If not click the link.
def test_run(app, client):
    with app.test_request_context():
        req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=1),
                           data=dict(filename=FILENAME),
                           location=url_for('jobs', node=1))
    sleep()

    ins = [
        '%s.log' % cst.JOBID,
        'USER_AGENT: Mozilla/5.0 (Windows',
        'ROBOTSTXT_OBEY: False',
        'COOKIES_ENABLED: False',
        'CONCURRENT_REQUESTS: 1',
        'DOWNLOAD_DELAY: 2',
        'CLOSESPIDER_TIMEOUT: 60',
        'CLOSESPIDER_PAGECOUNT: 10',
        'self.arg1: %s' % metadata['value']
    ]
    req_single_scrapyd(app, client, view='log',
                       kws=dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=cst.JOBID),
                       ins=ins)
    req_single_scrapyd(app, client, view='api',
                       kws=dict(node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID))


def test_run_fail(app, client):
    req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=1),
                       data={'filename': '%s_%s_%s.pickle' % (cst.PROJECT, cst.VERSION, cst.SPIDER)},
                       ins='Fail to schedule', set_to_second=True)


def test_schedule_xhr(app, client):
    req_single_scrapyd(app, client, view='schedule.xhr',
                       kws=dict(node=1, filename=FILENAME),
                       jskws=dict(status=cst.OK, jobid=cst.JOBID))
    req_single_scrapyd(app, client, view='api',
                       kws=dict(node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID))


def test_history(app, client):
    req_single_scrapyd(app, client, view='schedule.history', kws=dict(),
                       ins=['run_spider_history.log', ' -d arg1=%s' % metadata['value']])



# coding: utf-8
from flask import url_for

from tests.utils import cst, req, req_single_scrapyd


def test_node_reports_pass(app, client):
    with app.test_request_context():
        url_report = url_for('log', node=1, opt='report', project='PROJECT_PLACEHOLDER',
                             spider='SPIDER_PLACEHOLDER', job='JOB_PLACEHOLDER')
    ins = ["url_report: '%s'," % url_report, "start: '", "finish: '"]
    req(app, client, view='nodereports', kws=dict(node=1), ins=ins)
    req_single_scrapyd(app, client, view='nodereports', kws=dict(node=1), ins=ins)


def test_node_reports_fail(app, client):
    ins = ['<title>fail - ScrapydWeb</title>', '<h3>status_code: -1</h3>']
    req(app, client, view='nodereports', kws=dict(node=2), ins=ins)
    req_single_scrapyd(app, client, view='nodereports', kws=dict(node=1), ins=ins, set_to_second=True)


def test_cluster_reports(app, client):
    with app.test_request_context():
        url_servers = url_for('servers', node=2, opt='getreports', project=cst.PROJECT,
                              spider=cst.SPIDER, version_job=cst.JOBID)
        url_jobs = url_for('jobs', node=1)
        url_report = url_for('log', node=2, opt='report', project=cst.PROJECT,
                             spider=cst.SPIDER, job=cst.JOBID)
        url_redirect_to_clusterreports = url_for('clusterreports', node=1, project=cst.PROJECT,
                                                 spider=cst.SPIDER, job=cst.JOBID)
    ins = ['0 Reports of ////', '>Select a job</el-button>', url_jobs, 'selected_nodes: [],']
    nos = ['>Select nodes</el-button>']
    req(app, client, view='clusterreports', kws=dict(node=1), ins=ins)

    # Post from the servers page
    data = {
        '1': 'on',
        '2': 'on',
    }
    ins[0] = '%s Reports of /%s/%s/%s/' % (len(data), cst.PROJECT, cst.SPIDER, cst.JOBID)
    ins[-1] = 'selected_nodes: [1, 2],'
    ins.extend(nos)
    ins.append(url_servers)
    ins.append(url_report)
    kws = dict(node=2, project=cst.PROJECT, spider=cst.SPIDER, job=cst.JOBID)
    req(app, client, view='clusterreports', kws=kws, data=data, ins=ins)

    # Load metadata
    ins = ['<h1>Redirecting...</h1>', 'href="%s"' % url_redirect_to_clusterreports]
    req(app, client, view='clusterreports', kws=dict(node=1), ins=ins)



# coding: utf-8
import io
import os
import re

from flask import Blueprint, flash, redirect, render_template, request, send_from_directory, url_for
from logparser import parse
from werkzeug.utils import secure_filename

from ...vars import PARSE_PATH
from ..baseview import BaseView


ALLOWED_EXTENSIONS = {'log', 'txt'}
bp = Blueprint('parse', __name__, url_prefix='/')


@bp.route('/parse/source/<filename>')
def source(filename):
    return send_from_directory(PARSE_PATH, filename, mimetype='text/plain', cache_timeout=0)


class UploadLogView(BaseView):

    def __init__(self):
        super(UploadLogView, self).__init__()

        self.template = 'scrapydweb/parse.html'

    def dispatch_request(self, **kwargs):
        if self.POST:
            file = request.files.get('file')
            if not file:
                flash('No file selected', self.WARN)
                return redirect(request.url)

            if file.filename == '':
                flash('Filename not found', self.WARN)
                return redirect(request.url)

            if file.filename.rpartition('.')[-1] not in ALLOWED_EXTENSIONS:
                flash('Only file type of %s is supported' % ALLOWED_EXTENSIONS, self.WARN)
                return redirect(request.url)

            # Non-ASCII would be omitted and may set the filename as 'log' or 'txt'
            filename = secure_filename(file.filename)
            if filename in ALLOWED_EXTENSIONS:
                filename = '%s.%s' % (self.get_now_string(), filename)
            file.save(os.path.join(self.PARSE_PATH, filename))

            return redirect(url_for('.uploaded', node=self.node, filename=filename))
        else:
            url_parse_demo = url_for('.uploaded', node=self.node, filename='ScrapydWeb_demo.log')
            return render_template(self.template, node=self.node, url_parse_demo=url_parse_demo)


class UploadedLogView(BaseView):

    def __init__(self):
        super(UploadedLogView, self).__init__()

        self.filename = self.view_args['filename']
        self.template = 'scrapydweb/stats.html'

        self.text = ''
        self.project = ''
        self.spider = ''
        self.job = ''

    def dispatch_request(self, **kwargs):
        try:
            # Use io.open for compatibility with Python 2
            with io.open(os.path.join(self.PARSE_PATH, self.filename), encoding='utf-8', errors='ignore') as f:
                self.text = f.read()
        except Exception as err:
            return render_template(self.template_fail, node=self.node,
                                   alert="An error occurred when reading the uploaded logfile",
                                   text='%s\n%s' % (err.__class__.__name__, err))

        self.get_job_info()

        kwargs = dict(
            project=self.project,
            spider=self.spider,
            job=self.job,
            url_source=url_for('.source', filename=self.filename),
            # url_utf8=url_utf8, # To hide url_utf8 link in page http://127.0.0.1:5000/log/uploaded/ScrapydWeb_demo.log
        )
        kwargs.update(parse(self.text))
        # self.logger.debug("Parsed result: %s" % self.json_dumps(kwargs))
        return render_template(self.template, node=self.node, **kwargs)

    def get_job_info(self):
        # 2018-08-21 12:21:45 [scrapy.utils.log] INFO: Scrapy 1.5.0 started (bot: proxy)
        m_project = re.search(r'\(bot:\s(.+?)\)', self.text)
        self.project = m_project.group(1) if m_project else self.NA

        # 2018-08-21 12:21:45 [test] DEBUG: from_crawler
        m_spider = re.search(r'\[([^.]+?)\]\s+(?:DEBUG|INFO|WARNING|ERROR|CRITICAL)', self.text)
        self.spider = m_spider.group(1) if m_spider else self.NA

        # 'LOG_FILE': 'logs\\proxy\\test\\b2095ab0a4f911e8b98614dda9e91c2f.log',
        m_job = re.search(r'LOG_FILE.*?([\w-]+)\.(?:log|txt)', self.text)
        self.job = m_job.group(1) if m_job else (self.filename.rpartition('.')[0] or self.filename)



# coding: utf-8
import time

from flask import url_for
from apscheduler.schedulers.base import STATE_PAUSED, STATE_RUNNING

from scrapydweb import __version__
from scrapydweb.vars import SCHEDULER_STATE_DICT
from tests.utils import req


metadata = {}


def test_get_metadata(app, client):
    jskeys = [
        'jobs_per_page',  # tested
        'jobs_style',  #
        'id',
        'last_check_update_timestamp',  #
        'logparser_pid',
        'main_pid',
        'pageview',  #
        'password',
        'poll_pid',
        'scheduler_state',  #
        'tasks_per_page',  #
        'url_scrapydweb',
        'url_jobs',  #
        'url_schedule_task',  #
        'url_delete_task_result',  #
        'username',
        'version',  #
    ]
    with app.test_request_context():
        jskws = dict(
            version=__version__,
            url_jobs=url_for('jobs', node=1),  # '/1/jobs/',
            url_schedule_task=url_for('schedule.task', node=1),  # '/1/schedule/task/'
            # '/1/tasks/xhr/delete/1/1/'
            url_delete_task_result=url_for('tasks.xhr', node=1, action='delete', task_id=1, task_result_id=1)
        )
    __, js = req(app, client, view='metadata', kws=dict(node=1), jskeys=jskeys, jskws=jskws)
    print(js)
    metadata['last_check_update_timestamp'] = js['last_check_update_timestamp']


def test_last_check_update_timestamp(app, client):
    __, js = req(app, client, view='metadata', kws=dict(node=1),
                 jskws=dict(last_check_update_timestamp=metadata['last_check_update_timestamp'], pageview=1))
    assert time.time() - js['last_check_update_timestamp'] <= 3600 * 24 * 30


# :page-size="10"   size: '—',    long dash?!
def test_set_per_page(app, client):
    d = {10: u'—', 100: 'mini', 1000: 'mini'}
    for view in ['jobs', 'tasks']:
        for page in [10, 100, 1000, 100, 10]:
            __, js = req(app, client, view='metadata', kws=dict(node=1))
            another_per_page = 'tasks_per_page' if view == 'jobs' else 'jobs_per_page'
            before = js[another_per_page]

            ins = [":page-size='%s'" % page, "size: '%s'," % d[page]]
            req(app, client, view=view, kws=dict(node=1, style='database', per_page=page), ins=ins)
            req(app, client, view=view, kws=dict(node=1, style='database'), ins=ins)

            __, js = req(app, client, view='metadata', kws=dict(node=1))
            assert js['%s_per_page' % view] == page
            assert js[another_per_page] == before


def test_set_jobs_style(app, client):
    d = dict(database="Vue.extend(Main)", classic='class="table wrap"')
    for style in ['database', 'classic', 'database']:
        req(app, client, view='jobs', kws=dict(node=1, style=style), ins=d[style])
        req(app, client, view='metadata', kws=dict(node=1), jskws=dict(jobs_style=style))
        req(app, client, view='jobs', kws=dict(node=1, ui='mobile'), mobileui=True,
            ins='<table id="jobs", border="1">')
        req(app, client, view='jobs', kws=dict(node=1), ins=d[style])


def test_scheduler_state(app, client):
    for action, state in zip(['enable', 'disable', 'enable'], [STATE_RUNNING, STATE_PAUSED, STATE_RUNNING]):
        with app.test_request_context():
            if action == 'enable':
                scheduler_action_button = 'ENABLED'
                url_scheduler_action = url_for('tasks.xhr', node=1, action='disable')
            else:
                scheduler_action_button = 'DISABLED'
                url_scheduler_action = url_for('tasks.xhr', node=1, action='enable')
        req(app, client, view='tasks.xhr', kws=dict(node=1, action=action), ins=SCHEDULER_STATE_DICT[state])
        req(app, client, view='metadata', kws=dict(node=1), jskws=dict(scheduler_state=state))
        # ENABLED | DISABLED buttons
        req(app, client, view='tasks', kws=dict(node=1), ins=[scheduler_action_button, url_scheduler_action])



# coding: utf-8
from tests.utils import cst, req, upload_file_deploy


# def test_node_out_of_index(app, client):


# {'status': 'ok', 'pending': 0, 'running': 2, 'finished': 3}
def test_daemonstatus(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='daemonstatus'),
        jskws=dict(status=cst.OK), jskeys=['pending', 'running', 'finished'])


# def test_addversion(app, client):


# def test_schedule(app, client):


# {u'status': u'ok', u'prevstate': None, u'url': u'http://127.0.0.1:6800/cancel.json', u'status_code': 200}
# u'prevstate': running
def test_stop(app, client):
    upload_file_deploy(app, client, filename='demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT)

    req(app, client, view='api', kws=dict(node=1, opt='stop', project=cst.PROJECT, version_spider_job=cst.JOBID),
        nos='times', jskws=dict(status=cst.OK), jskeys='prevstate')


def test_forcestop(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='forcestop', project=cst.PROJECT, version_spider_job=cst.JOBID),
        jskws=dict(status=cst.OK, times=2, prevstate=None))


# {'status': 'ok', 'projects': ['demo']}
def test_listprojects(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='listprojects'),
        jskws=dict(status=cst.OK), jskeys='projects')


# {'status': 'ok', 'versions': []}
def test_listversions(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='listversions', project=cst.PROJECT),
        jskws=dict(status=cst.OK), jskeys='versions')


def listspiders(app, client, version):
    req(app, client, view='api', kws=dict(node=1, opt='listspiders', project=cst.PROJECT, version_spider_job=version),
        jskws=dict(status=cst.OK, spiders=cst.SPIDER))


def test_listspiders(app, client):
    for version in [cst.VERSION, cst.DEFAULT_LATEST_VERSION]:
        listspiders(app, client, version)


def test_listjobs(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='listjobs', project=cst.PROJECT),
        jskws=dict(status=cst.OK, url='listjobs.json'), jskeys=['pending', 'running', 'finished'])


# "message": "[WinError 32] 另一个程序正在使用此文件，进程无法访问。: 'eggs\\\\demo\\\\2018-01-01T01_01_01.egg'",
def test_delversion(app, client):
    kws = dict(node=1, opt='delversion', project=cst.PROJECT, version_spider_job=cst.VERSION)
    try:
        req(app, client, view='api', kws=kws, jskws=dict(status=cst.OK, url='delversion.json'))
    except AssertionError:
        req(app, client, view='api', kws=kws, jskws=dict(status=cst.ERROR, message='%s.egg' % cst.VERSION))


# "message": "[WinError 3] 系统找不到指定的路径。: 'eggs\\\\demo'",
def test_delproject(app, client):
    kws = dict(node=1, opt='delproject', project=cst.PROJECT)
    try:
        req(app, client, view='api', kws=kws, jskws=dict(status=cst.OK, url='delproject.json'))
    except AssertionError:
        req(app, client, view='api', kws=kws, jskws=dict(status=cst.ERROR, message=cst.PROJECT))


# After test_enable_logparser()
def test_liststats(app, client):
    # {'status': 'ok', 'details': {'pages': 'N/A', 'items': 'N/A', 'project': 'ScrapydWeb_demo', 'spider': 'N/A',
    # 'jobid': '2018-01-01T01_01_02', 'logparser_version': '0.8.0'}}
    # 'jobid': 'FAKE_JOBID'
    for jobid in [cst.JOBID, cst.FAKE_JOBID]:
        kws = dict(node=1, opt='liststats', project=cst.PROJECT, version_spider_job=jobid)
        __, js = req(app, client, view='api', kws=kws, jskeys=['status', 'details'])
        assert js['details']['project'] == cst.PROJECT
        if jobid == cst.FAKE_JOBID:
            assert js['details']['spider'] == cst.NA
        # else:
        #     assert js['details']['spider'] == cst.SPIDER
        assert js['details']['jobid'] == jobid
        assert 'pages' in js['details'] and 'items' in js['details']
        assert 'datas' not in js



# coding: utf-8
import json
import logging
import re
import time
import traceback

from ...common import get_now_string, get_response_from_view, handle_metadata
from ...models import Task, TaskResult, TaskJobResult, db
from ...utils.scheduler import scheduler


apscheduler_logger = logging.getLogger('apscheduler')

REPLACE_URL_NODE_PATTERN = re.compile(r'^/(\d+)/')
EXTRACT_URL_SERVER_PATTERN = re.compile(r'//(.+?:\d+)')


class TaskExecutor(object):

    def __init__(self, task_id, task_name, url_scrapydweb, url_schedule_task, url_delete_task_result,
                 url_listjobs, auth, selected_nodes):
        self.task_id = task_id
        self.task_name = task_name
        self.url_scrapydweb = url_scrapydweb
        self.url_schedule_task = url_schedule_task
        self.url_listjobs = url_listjobs
        self.url_delete_task_result = url_delete_task_result
        self.auth = auth
        self.data = dict(
            task_id=task_id,
            jobid='task_%s_%s' % (task_id, get_now_string(allow_space=False))
        )
        self.selected_nodes = selected_nodes
        self.task_result_id = None  # Be set in get_task_result_id()
        self.pass_count = 0
        self.fail_count = 0

        self.sleep_seconds_before_retry = 3
        self.nodes_to_retry = []
        self.logger = logging.getLogger(self.__class__.__name__)

    def main(self):
        self.get_task_result_id()
        for index, nodes in enumerate([self.selected_nodes, self.nodes_to_retry]):
            if not nodes:
                continue
            if index == 1:
                # https://apscheduler.readthedocs.io/en/latest/userguide.html#shutting-down-the-scheduler
                self.logger.warning("Retry task #%s (%s) on nodes %s in %s seconds",
                                    self.task_id, self.task_name, nodes, self.sleep_seconds_before_retry)
                time.sleep(self.sleep_seconds_before_retry)
                self.logger.warning("Retrying task #%s (%s) on nodes %s", self.task_id, self.task_name, nodes)
            for node in nodes:
                # Check the number of tasks of the same type running or pending

                result = self.schedule_task(node)
                if result:
                    if result['status'] == 'ok':
                        self.pass_count += 1
                    else:
                        self.fail_count += 1
                    self.db_insert_task_job_result(result)
        self.db_update_task_result()

    def get_task_result_id(self):
        # SQLite objects created in a thread can only be used in that same thread
        with db.app.app_context():
            task_result = TaskResult()
            task_result.task_id = self.task_id
            db.session.add(task_result)
            # db.session.flush()  # Get task_result.id before committing, flush() is part of commit()
            db.session.commit()
            # If directly use task_result.id later: Instance <TaskResult at 0x123> is not bound to a Session
            self.task_result_id = task_result.id
            self.logger.debug("Get new task_result_id %s for task #%s", self.task_result_id, self.task_id)




    def schedule_task(self, node):
        # TODO: Application was not able to create a URL adapter for request independent URL generation.
        # You might be able to fix this by setting the SERVER_NAME config variable.
        # with app.app_context():
        #     url_schedule_task = url_for('schedule.task', node=node)
        # http://127.0.0.1:5000/1/schedule/task/
        # /1/schedule/task/
        url_schedule_task = re.sub(REPLACE_URL_NODE_PATTERN, r'/%s/' % node, self.url_schedule_task)
        js = {}
        try:
            # assert '/1/' not in url_schedule_task, u"'故意出错'\r\n\"出错\"'故意出错'\r\n\"出错\""
            # assert False
            # time.sleep(10)
            js = get_response_from_view(url_schedule_task, auth=self.auth, data=self.data, as_json=True)
            assert js['status_code'] == 200 and js['status'] == 'ok', "Request got %s" % js
        except Exception as err:
            if node not in self.nodes_to_retry:
                apscheduler_logger.warning("Fail to execute task #%s (%s) on node %s, would retry later: %s",
                                           self.task_id, self.task_name, node, err)
                self.nodes_to_retry.append(node)
                return {}
            else:
                apscheduler_logger.error("Fail to execute task #%s (%s) on node %s, no more retries: %s",
                                         self.task_id, self.task_name, node, traceback.format_exc())
                js.setdefault('url', self.url_scrapydweb)  # '127.0.0.1:5000'
                js.setdefault('status_code', -1)
                js.setdefault('status', 'exception')
                js.setdefault('exception', traceback.format_exc())
        js.update(node=node)
        return js

    def db_insert_task_job_result(self, js):
        with db.app.app_context():
            if not TaskResult.query.get(self.task_result_id):
                apscheduler_logger.error("task_result #%s of task #%s not found", self.task_result_id, self.task_id)
                apscheduler_logger.warning("Discard task_job_result of task_result #%s of task #%s: %s",
                                           self.task_result_id, self.task_id, js)
                return
            task_job_result = TaskJobResult()
            task_job_result.task_result_id = self.task_result_id
            task_job_result.node = js['node']
            task_job_result.server = re.search(EXTRACT_URL_SERVER_PATTERN, js['url']).group(1)  # '127.0.0.1:6800'
            task_job_result.status_code = js['status_code']
            task_job_result.status = js['status']
            task_job_result.result = js.get('jobid', '') or js.get('message', '') or js.get('exception', '')
            db.session.add(task_job_result)
            db.session.commit()
            self.logger.info("Inserted task_job_result: %s", task_job_result)

    # https://stackoverflow.com/questions/13895176/sqlalchemy-and-sqlite-database-is-locked
    def db_update_task_result(self):
        with db.app.app_context():
            task = Task.query.get(self.task_id)
            task_result = TaskResult.query.get(self.task_result_id)
            if not task:
                apscheduler_logger.error("Task #%s not found", self.task_id)
                # if task_result:
                # '/1/tasks/xhr/delete/1/1/'
                url_delete_task_result = re.sub(r'/\d+/\d+/$', '/%s/%s/' % (self.task_id, self.task_result_id),
                                                self.url_delete_task_result)
                js = get_response_from_view(url_delete_task_result, auth=self.auth, data=self.data, as_json=True)
                apscheduler_logger.warning("Deleted task_result #%s [FAIL %s, PASS %s] of task #%s: %s",
                                           self.task_result_id, self.fail_count, self.pass_count, self.task_id, js)
                return
            if not task_result:
                apscheduler_logger.error("task_result #%s of task #%s not found", self.task_result_id, self.task_id)
                apscheduler_logger.warning("Failed to update task_result #%s [FAIL %s, PASS %s] of task #%s",
                                           self.task_result_id, self.fail_count, self.pass_count, self.task_id)
                return
            task_result.fail_count = self.fail_count
            task_result.pass_count = self.pass_count
            db.session.commit()
            self.logger.info("Inserted task_result: %s", task_result)


def execute_task(task_id):
    with db.app.app_context():
        task = Task.query.get(task_id)
        apscheduler_job = scheduler.get_job(str(task_id))
        if not task:
            apscheduler_job.remove()
            apscheduler_logger.error("apscheduler_job #{id} removed since task #{id} not exist. ".format(id=task_id))
        else:
            metadata = handle_metadata()
            username = metadata.get('username', '')
            password = metadata.get('password', '')
            url_delete_task_result = metadata.get('url_delete_task_result', '/1/tasks/xhr/delete/1/1/')
            task_executor = TaskExecutor(task_id=task_id,
                                         task_name=task.name,
                                         url_scrapydweb=metadata.get('url_scrapydweb', 'http://127.0.0.1:5000'),
                                         url_schedule_task=metadata.get('url_schedule_task', '/1/schedule/task/'),
                                         url_listjobs=metadata.get('url_listjobs',
                                                                   '/1/api/listjobs/'),
                                         url_delete_task_result=url_delete_task_result,
                                         auth=(username, password) if username and password else None,
                                         selected_nodes=json.loads(task.selected_nodes))
            try:
                task_executor.main()
            except Exception:
                apscheduler_logger.error(traceback.format_exc())



# coding: utf-8
import json

from flask import render_template, url_for

from ..baseview import BaseView


class NodeReportsView(BaseView):

    def __init__(self):
        super(NodeReportsView, self).__init__()

        self.url = url_for('jobs', node=self.node, listjobs='True')
        self.text = ''
        self.jobs = []
        self.pending_jobs = []
        self.running_jobs = []
        self.finished_jobs = []
        self.template = 'scrapydweb/node_reports.html'

    def dispatch_request(self, **kwargs):
        self.text = self.get_response_from_view(self.url, as_json=False)
        try:
            self.jobs = json.loads(self.text)
        except ValueError as err:
            self.logger.error("Fail to decode json from %s: %s", self.url, err)
            return self.text

        for job in self.jobs:
            if not job['start']:
                self.pending_jobs.append(job)
            else:
                if job['finish']:
                    self.finished_jobs.append(job)
                else:
                    self.running_jobs.append(job)

        if self.JOBS_FINISHED_JOBS_LIMIT > 0:
            self.finished_jobs = self.finished_jobs[::-1][:self.JOBS_FINISHED_JOBS_LIMIT]
        else:
            self.finished_jobs = self.finished_jobs[::-1]
        kwargs = dict(
            node=self.node,
            url=self.url,
            pending_jobs=self.pending_jobs,
            running_jobs=self.running_jobs,
            finished_jobs=self.finished_jobs,
            url_report=url_for('log', node=self.node, opt='report', project='PROJECT_PLACEHOLDER',
                               spider='SPIDER_PLACEHOLDER', job='JOB_PLACEHOLDER'),
            url_schedule=url_for('schedule', node=self.node),
        )
        return render_template(self.template, **kwargs)



# coding: utf-8
from flask import redirect, url_for

from .baseview import BaseView


class IndexView(BaseView):

    def __init__(self):
        super(IndexView, self).__init__()

    def dispatch_request(self, **kwargs):
        if self.SCRAPYD_SERVERS_AMOUNT == 1:
            if self.IS_MOBILE and not self.IS_IPAD:
                return redirect(url_for('jobs', node=self.node, ui='mobile'))
            else:
                return redirect(url_for('jobs', node=self.node, ui=self.UI))
        else:
            if self.USE_MOBILEUI or (self.IS_MOBILE and not self.IS_IPAD):
                return redirect(url_for('jobs', node=self.node, ui='mobile'))
            else:
                return redirect(url_for('servers', node=self.node, ui=self.UI))



# coding: utf-8
"""
source: https://github.com/scrapy/scrapyd-client
scrapyd-client/scrapyd_client/deploy.py
"""
import errno
import glob
import os
from shutil import copyfile
from subprocess import check_call
import sys
import tempfile

from flask import current_app as app
try:
    from six.moves.configparser import SafeConfigParser
except ImportError:
    from six.moves.configparser import RawConfigParser as SafeConfigParser


_SETUP_PY_TEMPLATE = """# Automatically created by: scrapydweb x scrapyd-client

from setuptools import setup, find_packages

setup(
    name         = 'project',
    version      = '1.0',
    packages     = find_packages(),
    entry_points = {'scrapy': ['settings = %(settings)s']},
)
"""


def get_config(sources):
    """Get Scrapy config file as a SafeConfigParser"""
    # sources = get_sources(use_closest)
    cfg = SafeConfigParser()
    cfg.read(sources)
    return cfg


def retry_on_eintr(func, *args, **kw):
    """Run a function and retry it while getting EINTR errors"""
    while True:
        try:
            return func(*args, **kw)
        except IOError as e:
            if e.errno != errno.EINTR:
                raise


def _build_egg(scrapy_cfg_path):
    cwd = os.getcwd()
   # If get_config() raise an error without executing os.chdir(cwd), would cause subsequent test cases
   # to raise TemplateNotFound when testing in Python 2 on Debian or macOS.
   # Debug: add print(environment.list_templates()) in flask/templating.py _get_source_fast() would show []
    try:
        os.chdir(os.path.dirname(scrapy_cfg_path))

        if os.path.exists('setup.py'):
            copyfile('setup.py', 'setup_backup.py')
        # lib/configparser.py: def get(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
        # projectname/scrapy.cfg: [settings] default = demo.settings
        settings = get_config(scrapy_cfg_path).get('settings', 'default')  # demo.settings
        _create_default_setup_py(settings=settings)

        d = tempfile.mkdtemp(prefix="scrapydweb-deploy-")
        o = open(os.path.join(d, "stdout"), "wb")
        e = open(os.path.join(d, "stderr"), "wb")
        retry_on_eintr(check_call, [sys.executable, 'setup.py', 'clean', '-a', 'bdist_egg', '-d', d],
                       stdout=o, stderr=e)
        egg = glob.glob(os.path.join(d, '*.egg'))[0]
        o.close()
        e.close()
    except:
        os.chdir(cwd)
        raise
    finally:
        os.chdir(cwd)

    return egg, d


def _create_default_setup_py(**kwargs):
    with open('setup.py', 'w') as f:
        content = _SETUP_PY_TEMPLATE % kwargs
        app.logger.debug('New setup.py')
        # app.logger.debug(content)
        f.write(content)



# coding: utf-8
import os

from scrapydweb import create_app
from scrapydweb.common import find_scrapydweb_settings_py
from scrapydweb.vars import SCRAPYDWEB_SETTINGS_PY
from scrapydweb.utils.check_app_config import check_app_config, check_email
from tests.utils import get_text, req
from tests.test_z_cleantest import test_cleantest as cleantest


# http://flask.pocoo.org/docs/1.0/tutorial/tests/#factory
def test_config():
    assert not create_app().testing
    assert create_app({'TESTING': True}).testing


def test_hello(client):
    response = client.get('/hello')
    assert get_text(response) == 'Hello, World!'


# def test_code_404(client):
    # response = client.get('/1/not-exist/')
    # assert response.status_code == 404 and 'Nothing Found' in get_text(response)


# How to test 500?
# def test_code_500(client):
    # response = client.get('/3/')
    # assert response.status_code == 500


def test_find_scrapydweb_settings_py():
    find_scrapydweb_settings_py(SCRAPYDWEB_SETTINGS_PY, os.getcwd())


def test_check_app_config(app, client):
    cleantest(app, client)

    # In conftest.py: ENABLE_LOGPARSER=False
    assert not os.path.exists(app.config['STATS_JSON_PATH'])

    # ['username:password@127.0.0.1:6800', ]
    app.config['SCRAPYD_SERVERS'] = app.config['_SCRAPYD_SERVERS']
    check_app_config(app.config)

    strings = []

    assert app.config['LOGPARSER_PID'] is None
    strings.append('logparser_pid: None')

    poll_pid = app.config['POLL_PID']
    if app.config.get('ENABLE_MONITOR', False):
        assert isinstance(poll_pid, int) and poll_pid > 0
    else:
        assert poll_pid is None
    strings.append('poll_pid: %s' % poll_pid)

    req(app, client, view='settings', kws=dict(node=1), ins=strings)
    assert not os.path.exists(app.config['STATS_JSON_PATH'])

    # Test ENABLE_MONITOR = False
    if app.config.get('ENABLE_MONITOR', False):
        app.config['ENABLE_MONITOR'] = False

        # ['username:password@127.0.0.1:6800', ]
        app.config['SCRAPYD_SERVERS'] = app.config['_SCRAPYD_SERVERS']
        check_app_config(app.config)

        assert app.config['LOGPARSER_PID'] is None
        assert app.config['POLL_PID'] is None
        req(app, client, view='settings', kws=dict(node=1), ins='poll_pid: None')

    # Test ENABLE_LOGPARSER = True, see test_enable_logparser()


def test_check_email_with_fake_account(app):
    with app.test_request_context():
        if not app.config.get('EMAIL_PASSWORD', ''):
            return

        app.config['EMAIL_USERNAME'] = 'username@qq.com'
        app.config['EMAIL_PASSWORD'] = 'password'
        app.config['EMAIL_SENDER'] = 'username@qq.com'
        app.config['EMAIL_RECIPIENTS'] = ['username@qq.com']
        try:
            check_email(app.config)
        except AssertionError:
            pass


def test_check_email_with_ssl_false(app):
    with app.test_request_context():
        if not app.config.get('EMAIL_PASSWORD', '') or not app.config.get('EMAIL_PASSWORD_'):
            return

        app.config['EMAIL_USERNAME'] = app.config['EMAIL_USERNAME_']
        app.config['EMAIL_PASSWORD'] = app.config['EMAIL_PASSWORD_']
        app.config['EMAIL_SENDER'] = app.config['EMAIL_SENDER_']
        app.config['EMAIL_RECIPIENTS'] = app.config['EMAIL_RECIPIENTS_']
        app.config['SMTP_SERVER'] = app.config['SMTP_SERVER_']
        app.config['SMTP_PORT'] = app.config['SMTP_PORT_']
        app.config['SMTP_OVER_SSL'] = app.config['SMTP_OVER_SSL_']
        app.config['SMTP_CONNECTION_TIMEOUT_'] = app.config['SMTP_CONNECTION_TIMEOUT_']

        check_email(app.config)


def test_scrapyd_group(app, client):
    req(app, client, view='servers', kws=dict(node=1), ins='Scrapyd-group')


def test_scrapyd_auth(app, client):
    req(app, client, view='settings', kws=dict(node=1), ins='u*e*n*m*:****56ab****')  # ('username', '123456abcdef')



# coding: utf-8
"""
REF: test_schedule_single_scrapyd.py
"""
import re

from flask import url_for
from six.moves.urllib.parse import unquote_plus
from tzlocal import get_localzone

from tests.utils import cst, req_single_scrapyd, sleep, upload_file_deploy


NODE = 1
NAME = u"""Chinese' "中文"""
VALUE = u"""Test' "测试"""
FILENAME = '%s_%s_%s.pickle' % (cst.PROJECT, cst.VERSION, cst.SPIDER)
FILENAME_DV = '%s_%s_%s.pickle' % (cst.PROJECT, 'default-the-latest-version', cst.SPIDER)
TITLE = '/'.join([cst.PROJECT, cst.VERSION, cst.SPIDER, cst.JOBID])
TITLE_DV = '/'.join([cst.PROJECT, cst.DEFAULT_LATEST_VERSION, cst.SPIDER, cst.JOBID])
metadata = {}
DATA = {}
DATA_DV = {}


# check POST default "action": "add"
def test_check_with_task(app, client):
    DATA.update(dict(
        project=cst.PROJECT,
        _version=cst.VERSION,
        spider=cst.SPIDER,
        jobid=cst.JOBID,
        USER_AGENT='iPhone',
        ROBOTSTXT_OBEY='True',
        COOKIES_ENABLED='False',
        CONCURRENT_REQUESTS='5',
        DOWNLOAD_DELAY='10',
        additional=("-d setting=CLOSESPIDER_PAGECOUNT=20\r\n"
                    u"-d arg1=%s\r\n"
                    "-d setting=CLOSESPIDER_TIMEOUT=120") % VALUE
    ))
    DATA.update(dict(
        year='2036',
        month='12',
        day='31',
        week='1',   # datetime.date(2036, 12, 31).isocalendar()[1]
        # day_of_week=['*'],
        day_of_week='mon-fri,sun',  # From browser: "*" | "*,mon-fri" | ""
        hour='2',
        minute='3',
        second='4',

        # start_date='',
        # end_date='',
        start_date='2019-01-01 00:00:00',
        end_date='2036-12-31 23:59:59',

        timezone='Asia/Shanghai',  # str(get_localzone())   'Asia/Shanghai'
        jitter='0',
        misfire_grace_time='600',
        coalesce='True',
        max_instances='1',

        task_id='0',
        action='add_fire',
        trigger='cron',
        name=NAME,
        replace_existing='True',
    ))
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=DATA,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))

    DATA_DV.update(dict(DATA))
    DATA_DV.update(_version=cst.DEFAULT_LATEST_VERSION)
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=DATA_DV,
                       jskws=dict(filename=FILENAME_DV),
                       nos="-d _version=")


def check_dumped_task_data(js, version=cst.VERSION, day_of_week='mon-fri,sun'):
    data = js['data']
    settings_dict = dict(
        ROBOTSTXT_OBEY=True,
        COOKIES_ENABLED=False,
        CONCURRENT_REQUESTS=5,
        DOWNLOAD_DELAY=10,
        CLOSESPIDER_PAGECOUNT=20,
        CLOSESPIDER_TIMEOUT=120
    )

    trigger_data = dict(
        year='2036',
        month='12',
        day='31',
        week='1',
        day_of_week=day_of_week,
        hour='2',
        minute='3',
        second='4',

        # start_date=None,
        # end_date=None,
        start_date='2019-01-01 00:00:00+08:00',
        end_date='2036-12-31 23:59:59+08:00',

        timezone='Asia/Shanghai',  # str(get_localzone())   'Asia/Shanghai'
        jitter=0,
    )
    assert data['id'] == js['task_id']
    assert isinstance(data['id'], int)
    # assert data['name'] is None
    assert data['name'] == NAME
    assert data['trigger'] == 'cron'
    assert data['create_time']
    assert data['update_time']

    assert data['project'] == cst.PROJECT
    assert data['version'] == version
    assert data['spider'] == cst.SPIDER
    assert data['jobid'] == cst.JOBID

    assert len(data['settings_arguments']) == 2
    assert data['settings_arguments']['arg1'] == VALUE
    assert len(data['settings_arguments']['setting']) == len(settings_dict) + 1
    assert 'USER_AGENT=Mozilla/5.0 (iPhone' in str(data['settings_arguments'])
    for k, v in settings_dict.items():
        assert '%s=%s' % (k, v) in data['settings_arguments']['setting']
    assert data['selected_nodes'] == [1]

    assert data['apscheduler_job']['trigger'] == trigger_data
    for k, v in trigger_data.items():
        if k in ['start_date', 'end_date']:
            assert v.startswith(data[k])
        else:
            assert data[k] == v

    assert data['apscheduler_job']['id'] == str(js['task_id'])  # str
    assert data['apscheduler_job']['kwargs'] == dict(task_id=js['task_id'])
    assert data['apscheduler_job']['misfire_grace_time'] == data['misfire_grace_time'] == 600
    assert data['apscheduler_job']['coalesce'] is True
    assert data['coalesce'] == 'True'
    assert data['apscheduler_job']['max_instances'] == data['max_instances'] == 1
    assert data['apscheduler_job']['next_run_time'] == "2036-12-31 02:03:04+08:00"


def test_enable_disable_scheduler(app, client):
    flash = "test flash in url"
    tip = "Click the DISABLED button to enable the scheduler for timer tasks"
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='enable'),
                       ins='STATE_RUNNING', nos='STATE_PAUSED')
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE), nos=tip)

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='disable'),
                       ins='STATE_PAUSED', nos='STATE_RUNNING')
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, flash=flash), ins=[flash, tip])

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='enable'),
                       ins='STATE_RUNNING', nos='STATE_PAUSED')
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE), nos=tip)


# You should be redirected automatically to target URL: <a href="/1/tasks/?flash=Add+task+%239+%28task_9%29+successfully
# {target} task #{task_id} ({task_name}) successfully
def test_run_with_task(app, client):
    # ScrapydWeb_demo.egg: custom_settings = {}, also output specific settings & arguments in the log
    upload_file_deploy(app, client, filename='ScrapydWeb_demo_no_request.egg', project=cst.PROJECT,
                       redirect_project=cst.PROJECT)

    with app.test_request_context():
        metadata['location'] = url_for('tasks', node=NODE)
    # 'schedule.check' in test_check_with_task()
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  location=metadata['location'])
    sleep()
    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    next_run_time = m.group(2)
    print("task_id: %s" % task_id)
    print("next_run_time: %s" % next_run_time)
    metadata['task_id'] = task_id
    metadata['next_run_time'] = next_run_time

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    check_dumped_task_data(js)
    text, __ = req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE))
    jobid = re.search(r'/(task_%s_[\w-]+?)/' % task_id, text).group(1)  # extract jobid from url_stats in tasks
    print("jobid: %s" % jobid)
    metadata['jobid'] = jobid

    ins = [
        '%s.log' % jobid,
        'USER_AGENT: Mozilla/5.0 (iPhone',
        'ROBOTSTXT_OBEY: True',
        'COOKIES_ENABLED: False',
        'CONCURRENT_REQUESTS: 5',
        'DOWNLOAD_DELAY: 10',
        'CLOSESPIDER_TIMEOUT: 120',
        'CLOSESPIDER_PAGECOUNT: 20',
        (u'self.arg1: %s' % VALUE).replace("'", '&#39;').replace('"', '&#34;')
    ]
    # In utf8 page: <div id="log">  [test] DEBUG: self.arg1: Test&#39; &#34;测试   </pre>
    # https://stackoverflow.com/questions/2087370/decode-html-entities-in-python-string
    req_single_scrapyd(app, client, view='log',
                       kws=dict(node=NODE, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=jobid),
                       ins=ins)
    req_single_scrapyd(app, client, view='api',
                       kws=dict(node=NODE, opt='forcestop', project=cst.PROJECT, version_spider_job=jobid))


def test_check_tasks(app, client):
    task_id = metadata['task_id']
    jobid = metadata['jobid']
    with app.test_request_context():
        metadata['url_stats'] = url_for('log', node=NODE, opt='stats',
                                        project=cst.PROJECT, spider=cst.SPIDER, job=jobid)
        metadata['url_tasks'] = url_for('tasks', node=NODE)
        metadata['url_task_results'] = url_for('tasks', node=NODE, task_id=task_id)
        metadata['url_pause'] = url_for('tasks.xhr', node=NODE, action='pause', task_id=task_id)
        metadata['url_resume'] = url_for('tasks.xhr', node=NODE, action='resume', task_id=task_id)
        metadata['url_fire'] = url_for('tasks.xhr', node=NODE, action='fire', task_id=task_id)
        metadata['url_stop'] = url_for('tasks.xhr', node=NODE, action='remove', task_id=task_id)
        metadata['url_delete'] = url_for('tasks.xhr', node=NODE, action='delete', task_id=task_id)
        metadata['url_edit'] = url_for('schedule', node=NODE, task_id=task_id)

    ins_keys = ['url_pause', 'url_fire', 'url_stop', 'url_edit', 'url_stats', 'url_task_results']
    nos_keys = ['url_resume', 'url_delete']
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[metadata[k] for k in ins_keys] + ["prev_run_result: '%s'," % jobid[-19:]],
                       nos=[metadata[k] for k in nos_keys])

    next_run_time_valid = "next_run_time: '2036-12-31 02:03:04+08:00',"
    next_run_time_invalid = "next_run_time: 'Click DISABLED button first. ',"
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=next_run_time_valid, nos=next_run_time_invalid)

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='disable'),
                       ins='STATE_PAUSED', nos='STATE_RUNNING')
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=next_run_time_invalid, nos=next_run_time_valid)

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='enable'),
                       ins='STATE_RUNNING', nos='STATE_PAUSED')
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=next_run_time_valid, nos=next_run_time_invalid)


def test_check_task_results_with_job(app, client):
    task_id = metadata['task_id']
    metadata['server'] = app.config['SCRAPYD_SERVERS'][0]  # when fail: [node 1] 127.0.0.1:5000
    text, __ = req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                                  ins=[TITLE, metadata['url_tasks'], '[node 1] %s' % metadata['server'],
                                       "status_code: 200,", "status: 'ok',", metadata['url_stats'],
                                       "result: '%s'," % metadata['jobid'], ":total='1'"],
                                  nos=[metadata['url_task_results'], 'label="Pass count"', 'label="Server"'])

    # in the task results page: url_action: '/1/tasks/xhr/delete/5/10/',
    task_result_id = int(re.search(r'%s(\d+)/' % metadata['url_delete'], text).group(1))
    print("task_result_id: %s" % task_result_id)
    metadata['task_result_id'] = task_result_id

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id),
                       jskws=dict(ids=[task_result_id]))

    with app.test_request_context():
        metadata['url_delete_task_result'] = url_for('tasks.xhr', node=NODE, action='delete',
                                                     task_id=task_id, task_result_id=task_result_id)
    assert metadata['url_delete_task_result'] in text


def test_check_task_job_results(app, client):
    task_id = metadata['task_id']
    task_result_id = metadata['task_result_id']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr',
                                kws=dict(node=NODE, action='list', task_id=task_id, task_result_id=task_result_id))
    assert len(js['ids']) == 1

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=task_result_id),
                       ins=[TITLE, metadata['url_tasks'], metadata['url_task_results'],
                            'label="Server"', 'label="Node"', metadata['server'],
                            "status_code: 200,", "status: 'ok',", metadata['url_stats'],
                            "result: '%s'," % metadata['jobid'], ":total='1'"],
                       nos=metadata['url_delete_task_result'])


def test_check_task_not_exist(app, client):
    # edit_task
    req_single_scrapyd(app, client, view='schedule', kws=dict(node=NODE, task_id=cst.BIGINT),
                       ins=["fail - ScrapydWeb", "Task #%s not found" % cst.BIGINT])

    # dump
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=cst.BIGINT),
                       jskws=dict(data=None, status=cst.ERROR, message="Task #%s not found" % cst.BIGINT))

    # fire
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='fire', task_id=cst.BIGINT),
                       jskws=dict(status=cst.ERROR, message="apscheduler_job #%s not found" % cst.BIGINT),
                       nos=metadata['url_task_results'])

    # task_results
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=cst.BIGINT),
                       ins=["fail - ScrapydWeb", "Task #%s not found" % cst.BIGINT])

    # task_job_results
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=cst.BIGINT, task_result_id=1),
                       ins=["fail - ScrapydWeb", "Task #%s not found" % cst.BIGINT])


# TasksXhrView fire|pause|resume|remove|delete
def test_task_xhr_fire(app, client):
    task_id = metadata['task_id']
    # Note that the url_stats in the tasks page would be changed after a fire
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='fire', task_id=task_id),
                       jskws=dict(status=cst.OK, tip="Reload this page", url_jump=metadata['url_task_results']))
    sleep()
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
    assert len(js['ids']) == 2
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id), ins=":total='2'")


# apscheduler_job #1 after 'pause': task_1 (trigger:
# cron[year='*', month='*', day='*', week='*', day_of_week='*', hour='*', minute='0', second='0'], paused)
def test_task_xhr_pause(app, client):
    task_id = metadata['task_id']
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='pause', task_id=task_id),
                       jskws=dict(status=cst.OK, tip=", paused)"))

    ins_keys = ['url_resume', 'url_stop', 'url_edit', 'url_task_results']  # without url_stats
    nos_keys = ['url_pause', 'url_fire', 'url_delete']
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[metadata[k] for k in ins_keys],
                       nos=[metadata[k] for k in nos_keys])
    # fire a paused task
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='fire', task_id=task_id),
                       jskws=dict(status=cst.ERROR, message="resume it first"),
                       nos=['url_jump', metadata['url_task_results']])


# task_1 (trigger: cron[year='*', month='*', day='*', week='*', day_of_week='*', hour='*', minute='0', second='0'],
# next run at: 2019-01-01 00:00:01 CST)
def test_task_xhr_resume(app, client):
    task_id = metadata['task_id']
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='resume', task_id=task_id),
                       jskws=dict(status=cst.OK, tip="next run at:"))

    ins_keys = ['url_pause', 'url_fire', 'url_stop', 'url_edit', 'url_task_results']  # without url_stats
    nos_keys = ['url_resume', 'url_delete']
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[metadata[k] for k in ins_keys],
                       nos=[metadata[k] for k in nos_keys])


# 'Stop' button to 'remove' an apscheduler_job
def test_task_xhr_remove(app, client):
    task_id = metadata['task_id']
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='remove', task_id=task_id),
                       jskws=dict(status=cst.OK, tip="apscheduler_job #%s after 'remove': None" % task_id))
    ins_keys = ['url_delete', 'url_edit', 'url_task_results']  # without url_stats
    nos_keys = ['url_pause', 'url_resume', 'url_fire', 'url_stop']
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[metadata[k] for k in ins_keys],
                       nos=[metadata[k] for k in nos_keys])

    # Make sure the task results is not deleted
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
    assert len(js['ids']) == 2
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id), ins=":total='2'")
    # handle_apscheduler_job which is not exist
    for action in ['pause', 'resume', 'remove']:
        req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action=action, task_id=task_id),
                           jskws=dict(status=cst.ERROR, message="apscheduler_job #%s not found" % task_id))
    # dump task data without apscheduler_job
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id),
                                jskws=dict(tip="apscheduler_job #%s not found" % task_id))
    assert js['data']['apscheduler_job'] is None


def test_task_xhr_delete_a_task_result(app, client):
    task_id = metadata['task_id']
    task_result_id = metadata['task_result_id']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
    assert len(js['ids']) == 2 and task_result_id in js['ids']
    req_single_scrapyd(app, client, view='tasks.xhr',
                       kws=dict(node=NODE, action='delete', task_id=task_id, task_result_id=task_result_id),
                       jskws=dict(status=cst.OK, tip="task_result #%s deleted" % task_result_id))
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
    assert len(js['ids']) == 1 and task_result_id not in js['ids']

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE), ins=metadata['url_task_results'])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=":total='1'", nos=metadata['url_delete_task_result'])

    # delete a task result which is not exist
    req_single_scrapyd(app, client, view='tasks.xhr',
                       kws=dict(node=NODE, action='delete', task_id=task_id, task_result_id=task_result_id),
                       jskws=dict(status=cst.ERROR, message="task_result #%s not found" % task_result_id))


# remove apscheduler_job before deleting task
def test_task_xhr_delete_a_task_with_job(app, client):
    task_id = metadata['task_id']
    # 'schedule.check' in test_check_with_task()
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME_DV),
                                  location=metadata['location'])
    new_task_id = int(re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text)).group(1))
    assert new_task_id - task_id == 1

    sleep()  # Wait until the first execution finish
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=new_task_id))
    check_dumped_task_data(js, version=cst.DEFAULT_LATEST_VERSION)

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert new_task_id in js['ids']
    tip = "apscheduler_job #{id} removed. Task #{id} deleted".format(id=new_task_id)
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=new_task_id),
                       jskws=dict(status=cst.OK, tip=tip))
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert new_task_id not in js['ids']

    message = "apscheduler_job #{id} not found. Task #{id} not found. ".format(id=new_task_id)
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=new_task_id),
                                jskws=dict(status=cst.ERROR, message=message))
    assert js['data'] is None


def test_edit_task(app, client):
    task_id = metadata['task_id']
    ins = [
        "jobid: '%s'," % cst.JOBID,
        "USER_AGENT: 'iPhone',",
        "ROBOTSTXT_OBEY: 'True',",
        "COOKIES_ENABLED: 'False',",
        "CONCURRENT_REQUESTS: '5',",
        "DOWNLOAD_DELAY: '10',",
        ("additional: '-d setting=CLOSESPIDER_PAGECOUNT=20\\r\\n-d setting=CLOSESPIDER_TIMEOUT=120\\r\\n"
         u"-d arg1=%s\\r\\n',") % (VALUE.replace("'", "\\'")),  # In HTML:  additional: '-d arg1=Test\' "测试\r\n',

        "task_id: %s," % task_id,
        "replace_existing: 'True',",
        "action: 'add_fire',",
        "trigger: 'cron',",
        u"name: '%s - edit'," % (NAME.replace("'", "\\'")),  # In HTML:  name: 'Chinese\' "中文 - edit',

        "year: '2036',",
        "month: '12',",
        "day: '31',",
        "week: '1',",
        "day_of_week: ['mon-fri', 'sun'],",
        "hour: '2',",
        "minute: '3',",
        "second: '4',",
        "start_date: '2019-01-01 00:00:00',",
        "end_date: '2036-12-31 23:59:59',",

        "timezone: 'Asia/Shanghai',",
        "jitter: 0,",
        "misfire_grace_time: 600,",
        "coalesce: 'True',",
        "max_instances: 1,",

        "expandSettingsArguments: true,",
        "expandTimerTask: true,",
        "expandTimerTaskMoreSettings: false,",
    ]
    # http://127.0.0.1:5000/1/schedule/?task_id=1
    req_single_scrapyd(app, client, view='schedule', kws=dict(node=NODE, task_id=task_id), ins=ins)


# POST data contains "task_id": "1", "replace_existing": "True",
def test_edit_to_update_a_task(app, client):
    day_of_week = 'mon-fri'
    task_id = metadata['task_id']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    create_time = js['data']['create_time']
    update_time = js['data']['update_time']

    data = dict(DATA)
    data.update(task_id=task_id, day_of_week=day_of_week)  # modify day_of_week only
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  location=metadata['location'])
    assert int(re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text)).group(1)) == task_id
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids'] and (task_id + 1) not in js['ids']

    sleep()
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['create_time'] == create_time
    assert js['data']['update_time'] > update_time
    check_dumped_task_data(js, day_of_week=day_of_week)

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id + 1),
                       jskws=dict(data=None, status=cst.ERROR, message="Task #%s not found" % (task_id + 1)))
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "day_of_week: '%s'," % day_of_week], nos="id: %s," % (task_id + 1))


def test_edit_to_update_a_task_fail(app, client):
    day_of_week = '*'
    second = '10/*'  # ERROR INPUT
    task_id = metadata['task_id']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    create_time = js['data']['create_time']
    update_time = js['data']['update_time']

    data = dict(DATA)
    data.update(task_id=task_id, day_of_week=day_of_week, second=second)  # modify day_of_week and second
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  ins=[second, "re-edit task #%s" % task_id, "fail - ScrapydWeb"])
    # re-edit task #1
    task_id_ = re.search(r're-edit task #(\d+)', text).group(1)
    assert int(task_id_) == task_id
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids'] and (task_id + 1) not in js['ids']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['create_time'] == create_time
    assert js['data']['update_time'] == update_time
    check_dumped_task_data(js, day_of_week='mon-fri')  # Updated day_of_week in test_edit_to_update_a_task()

    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id + 1),
                       jskws=dict(data=None, status=cst.ERROR, message="Task #%s not found" % (task_id + 1)))
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "day_of_week: '%s'," % 'mon-fri'], nos="id: %s," % (task_id + 1))


# POST data contains "task_id": "1", "replace_existing": "False",
def test_edit_to_new_a_task(app, client):
    day_of_week = '*'
    task_id = metadata['task_id']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    create_time = js['data']['create_time']
    update_time = js['data']['update_time']

    data = dict(DATA)
    data.update(task_id=task_id, replace_existing='False', day_of_week=day_of_week)
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  location=metadata['location'])
    new_task_id = int(re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text)).group(1))
    print("new_task_id: %s" % new_task_id)
    # assert new_task_id == task_id + 1
    # For compatibility with postgresql, though test_task_xhr_delete_a_task_with_job is executed before
    # https://stackoverflow.com/questions/9984196/postgresql-gapless-sequences
    assert new_task_id > task_id
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids'] and new_task_id in js['ids']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['create_time'] == create_time
    assert js['data']['update_time'] == update_time
    check_dumped_task_data(js, day_of_week='mon-fri')

    sleep()
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=new_task_id))
    check_dumped_task_data(js, day_of_week=day_of_week)

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "day_of_week: 'mon-fri',",
                            "id: %s," % new_task_id, "day_of_week: '%s'," % day_of_week])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=new_task_id),
                       jskws=dict(status=cst.OK, tip="Task #%s deleted" % new_task_id))


# POST data contains "task_id": "1", "replace_existing": "False",
def test_edit_to_new_a_task_fail(app, client):
    day_of_week = '*'
    second = '10/*'  # ERROR INPUT
    task_id = metadata['task_id']
    data = dict(DATA)
    data.update(task_id=task_id, replace_existing='False', day_of_week=day_of_week, second=second)
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  ins=[day_of_week, second, "re-edit task #", "fail - ScrapydWeb"])
    # re-edit task #2
    new_task_id = int(re.search(r're-edit task #(\d+)', text).group(1))
    print("new_task_id: %s" % new_task_id)
    # For compatibility with postgresql
    assert new_task_id > task_id
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids'] and new_task_id in js['ids']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    check_dumped_task_data(js, day_of_week='mon-fri')
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=new_task_id))
    assert js['data']['apscheduler_job'] is None

    with app.test_request_context():
        url_pause = url_for('tasks.xhr', node=NODE, action='pause', task_id=new_task_id)
        url_resume = url_for('tasks.xhr', node=NODE, action='resume', task_id=new_task_id)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "day_of_week: 'mon-fri',",
                            "id: %s," % new_task_id, "day_of_week: '%s'," % day_of_week, "second: '%s'," % second],
                       nos=[url_pause, url_resume])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=new_task_id),
                       jskws=dict(status=cst.OK, tip="Task #%s deleted" % new_task_id))


def test_task_xhr_delete_a_task(app, client):
    task_id = metadata['task_id']

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='remove', task_id=task_id),
                       jskws=dict(status=cst.OK, tip="apscheduler_job #%s after 'remove': None" % task_id))

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id),
                       jskws=dict(status=cst.OK,
                                  tip="apscheduler_job #{id} not found. Task #{id} deleted".format(id=task_id)))
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id not in js['ids']

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE), nos=metadata['url_task_results'])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=["fail - ScrapydWeb", "Task #%s not found" % task_id])
    # delete a task which is not exist
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id),
                       jskws=dict(status=cst.ERROR, message="Task #%s not found" % task_id))


# Unrecognized expression "10/*" for field "second"
def test_add_task_fail(app, client):
    data_dv = dict(DATA_DV)
    data_dv.update(second='10/*')
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data_dv,
                       jskws=dict(filename=FILENAME_DV))
    ins = ["Unrecognized expression", "10/*", "re-edit task #", "kwargs for", "task_data for", "fail - ScrapydWeb"]
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME_DV),
                                  ins=ins)
    # re-edit task #1
    task_id = int(re.search(r're-edit task #(\d+)', text).group(1))
    print("task_id: %s" % task_id)
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']

    with app.test_request_context():
        url_pause = url_for('tasks.xhr', node=NODE, action='pause', task_id=task_id)
        url_resume = url_for('tasks.xhr', node=NODE, action='resume', task_id=task_id)

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["second: '10/*',", "status: 'Finished',", "prev_run_result: 'N/A',",
                            "next_run_time: '%s'," % cst.NA, "fail_times: 0,", "run_times: 0,"],
                       nos=[url_resume, url_pause])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=['label="Fail count"', 'label="Pass count"', ":total='0'"])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


# check POST  "action": "add"
def test_add_task_action_add(app, client):
    data_dv = dict(DATA_DV)
    data_dv.update(action='add', hour='*', minute='*', second='*')
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data_dv,
                       jskws=dict(filename=FILENAME_DV))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME_DV),
                                  location=metadata['location'])

    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    print("task_id: %s" % task_id)
    next_run_time = m.group(2)
    print("next_run_time: %s" % next_run_time)
    assert next_run_time == "2036-12-31 00:00:00+08:00"

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['apscheduler_job']['next_run_time'] == next_run_time

    with app.test_request_context():
        url_pause = url_for('tasks.xhr', node=NODE, action='pause', task_id=task_id)
        url_resume = url_for('tasks.xhr', node=NODE, action='resume', task_id=task_id)

    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[url_pause, "prev_run_result: 'N/A',", "next_run_time: '%s'," % next_run_time,
                            "fail_times: 0,", "run_times: 0,"],
                       nos=url_resume)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=[TITLE_DV, 'label="Pass count"', ":total='0'"])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=cst.BIGINT),
                       ins=[TITLE_DV, 'label="Server"', ":total='0'"])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


# check POST  "action": "add_pause"
def test_add_task_action_pause(app, client):
    data_dv = dict(DATA_DV)
    data_dv.update(action='add_pause', hour='1-3', minute='4-6', second='7-9')
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data_dv,
                       jskws=dict(filename=FILENAME_DV))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME_DV),
                                  location=metadata['location'])

    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    print("task_id: %s" % task_id)
    next_run_time = m.group(2)
    print("next_run_time: %s" % next_run_time)
    assert next_run_time == cst.NA

    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
    assert task_id in js['ids']
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['apscheduler_job']['next_run_time'] is None

    with app.test_request_context():
        url_pause = url_for('tasks.xhr', node=NODE, action='pause', task_id=task_id)
        url_resume = url_for('tasks.xhr', node=NODE, action='resume', task_id=task_id)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=[url_resume, "prev_run_result: 'N/A',", "next_run_time: '%s'," % cst.NA,
                            "fail_times: 0,", "run_times: 0,"],
                       nos=url_pause)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=[TITLE_DV, 'label="Pass count"', ":total='0'"])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=cst.BIGINT),
                       ins=[TITLE_DV, 'label="Server"', ":total='0'"])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


def test_add_task_button(app, client):
    ins = [
        "expandSettingsArguments: false,",
        "expandTimerTask: true,",
        "expandTimerTaskMoreSettings: false,",

        "jobid: '',",
        "USER_AGENT: '',",
        "ROBOTSTXT_OBEY: '',",
        "COOKIES_ENABLED: '',",
        "CONCURRENT_REQUESTS: '',",
        "DOWNLOAD_DELAY: '',",
        "additional: '-d setting=CLOSESPIDER_TIMEOUT=60\\r\\n-d setting=CLOSESPIDER_PAGECOUNT=10\\r\\n-d arg1=val1'",

        "task_id: 0,",
        "replace_existing: 'True',",
        "action: 'add_fire',",
        "trigger: 'cron',",
        "name: '',",

        "year: '*',",
        "month: '*',",
        "day: '*',",
        "week: '*',",
        "day_of_week: ['*'],",
        "hour: '*',",
        "minute: '0',",
        "second: '0',",
        "start_date: '',",
        "end_date: '',",

        "timezone: '%s'," % str(get_localzone()),
        "jitter: 0,",
        "misfire_grace_time: 600,",
        "coalesce: 'True',",
        "max_instances: 1,",
    ]
    req_single_scrapyd(app, client, view='schedule', kws=dict(node=NODE, add_task='True'), ins=ins)


def test_add_task_with_default_values(app, client):
    data = dict(DATA)
    for k in data.keys():
        if k in ['jitter', 'misfire_grace_time', 'max_instances']:
            data[k] = 'invalid int'
        elif k not in ['project', '_version', 'spider', 'trigger']:  # if not request.form.get('trigger'): return
            data[k] = ''
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  location=metadata['location'])
    sleep()
    task_id = int(re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text)).group(1))
    print("task_id: %s" % task_id)
    __, js = req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['settings_arguments'] == {'setting': []}
    assert js['data']['selected_nodes'] == [1]
    assert js['data']['timezone'] is None
    assert js['data']['apscheduler_job']['misfire_grace_time'] == 600
    assert js['data']['apscheduler_job']['coalesce'] is True
    assert js['data']['apscheduler_job']['max_instances'] == 1
    assert js['data']['apscheduler_job']['name'] == 'task_%s' % task_id
    assert ':00:00' in js['data']['apscheduler_job']['next_run_time']
    for k, v in js['data']['apscheduler_job']['trigger'].items():
        if k in ['start_date', 'end_date']:
            assert v is None
        elif k in ['minute', 'second']:
            assert v == '0'
        elif k == 'jitter':
            assert v == 0
        elif k == 'timezone':
            assert v == str(get_localzone())
        else:
            assert v == '*'
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


def test_execute_task_fail(app, client):
    data = dict(DATA)
    # set_to_second
    req_single_scrapyd(app, client, view='schedule.check', kws=dict(node=NODE), data=data, set_to_second=True,
                       jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req_single_scrapyd(app, client, view='schedule.run', kws=dict(node=NODE),
                                  data=dict(filename=FILENAME),
                                  location=metadata['location'])
    task_id = int(re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text)).group(1))
    print("task_id: %s" % task_id)
    # For compatibility with postgresql
    metadata['task_id'] = task_id
    sleep(2)
    # The first execution has not finished yet
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "prev_run_result: 'FAIL 0, PASS 0',"])
    text, __ = req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                                  ins=["fail_count: 0,", "pass_count: 0,", ":total='1'"])
    # in the task results page: url_action: '/1/tasks/xhr/delete/5/10/',
    with app.test_request_context():
        url_delete = url_for('tasks.xhr', node=NODE, action='delete', task_id=task_id)
    task_result_id = int(re.search(r'%s(\d+)/' % url_delete, text).group(1))
    print("task_result_id: %s" % task_result_id)
    sleep(28)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE),
                       ins=["id: %s," % task_id, "prev_run_result: 'FAIL 1, PASS 0',"])
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                       ins=["status_code: -1,", "status: 'error',", "Max retries exceeded", ":total='1'"],
                       nos="node: %s," % NODE)
    req_single_scrapyd(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=task_result_id),
                       ins=["node: %s," % NODE, "status_code: -1,", "status: 'error',",
                            "Max retries exceeded", ":total='1'"])
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


def test_history(app, client):
    task_id = metadata['task_id']
    next_run_time = metadata['next_run_time']
    req_single_scrapyd(app, client, view='tasks.history', kws=dict(),
                       ins=["timer_tasks_history.log",
                            "Add task #%s (%s) successfully" % (task_id, NAME),
                            "Added job_instance:",
                            "next run at %s" % next_run_time,
                            '"id": "%s",' % task_id,
                            '"name": "%s",' % (NAME.replace('"', '\\"')),  # "name": "Chinese' \"中文",
                            '"next_run_time": "datetime.datetime(',
                            '"trigger": "<CronTrigger',
                            "Fail to execute task #%s (%s) on node 1, would retry later" % (task_id, NAME),
                            "Max retries exceeded",
                            "Fail to execute task #%s (%s) on node 1, no more retries" % (task_id, NAME),
                            "assert js['status_code'] == 200 and js['status'] == 'ok'",
                            "Task #%s deleted" % task_id,
                            ])



# coding: utf-8
import argparse
import logging
import os
from shutil import copyfile
import sys

from flask import request

# from . import create_app  # --debug: ImportError: cannot import name 'create_app'
from scrapydweb import create_app
from scrapydweb.__version__ import __description__, __version__
from scrapydweb.common import authenticate, find_scrapydweb_settings_py, handle_metadata, handle_slash
from scrapydweb.vars import ROOT_DIR, SCRAPYDWEB_SETTINGS_PY, SCHEDULER_STATE_DICT, STATE_PAUSED, STATE_RUNNING
from scrapydweb.utils.check_app_config import check_app_config


logger = logging.getLogger(__name__)
apscheduler_logger = logging.getLogger('apscheduler')

STAR = '\n%s\n' % ('*' * 100)
DEFAULT_SETTINGS_PY_PATH = os.path.join(ROOT_DIR, 'default_settings.py')


def main():
    apscheduler_logger.setLevel(logging.ERROR)  # To hide warning logging in scheduler.py until app.run()
    main_pid = os.getpid()
    logger.info("ScrapydWeb version: %s", __version__)
    logger.info("Use 'scrapydweb -h' to get help")
    logger.info("Main pid: %s", main_pid)
    logger.debug("Loading default settings from %s", handle_slash(DEFAULT_SETTINGS_PY_PATH))
    app = create_app()
    handle_metadata('main_pid', main_pid)  # In handle_metadata(): with db.app.app_context():
    app.config['MAIN_PID'] = main_pid
    app.config['DEFAULT_SETTINGS_PY_PATH'] = DEFAULT_SETTINGS_PY_PATH
    app.config['SCRAPYDWEB_SETTINGS_PY_PATH'] = os.path.join(os.getcwd(), SCRAPYDWEB_SETTINGS_PY)
    load_custom_settings(app.config)

    args = parse_args(app.config)
    # "scrapydweb -h" ends up here
    update_app_config(app.config, args)
    try:
        check_app_config(app.config)
    except AssertionError as err:
        logger.error("Check app config fail: ")
        sys.exit(u"\n{err}\n\nCheck and update your settings in {path}\n".format(
                 err=err, path=handle_slash(app.config['SCRAPYDWEB_SETTINGS_PY_PATH'])))

    # https://stackoverflow.com/questions/34164464/flask-decorate-every-route-at-once
    @app.before_request
    def require_login():
        if app.config.get('ENABLE_AUTH', False):
            auth = request.authorization
            USERNAME = str(app.config.get('USERNAME', ''))  # May be 0 from config file
            PASSWORD = str(app.config.get('PASSWORD', ''))
            if not auth or not (auth.username == USERNAME and auth.password == PASSWORD):
                return authenticate()

    # MUST be commented out for released version
    # https://stackoverflow.com/questions/34066804/disabling-caching-in-flask
    # @app.after_request
    # def add_header(r):
        # r.headers['Pragma'] = 'no-cache'
        # r.headers['Expires'] = '0'
        # r.headers['Cache-Control'] = 'public, max-age=0'
        # return r

    @app.context_processor
    def inject_variable():
        SCRAPYD_SERVERS = app.config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
        SCRAPYD_SERVERS_PUBLIC_URLS = app.config.get('SCRAPYD_SERVERS_PUBLIC_URLS', None)
        return dict(
            SCRAPYD_SERVERS=SCRAPYD_SERVERS,
            SCRAPYD_SERVERS_AMOUNT=len(SCRAPYD_SERVERS),
            SCRAPYD_SERVERS_GROUPS=app.config.get('SCRAPYD_SERVERS_GROUPS', []) or [''],
            SCRAPYD_SERVERS_AUTHS=app.config.get('SCRAPYD_SERVERS_AUTHS', []) or [None],
            SCRAPYD_SERVERS_PUBLIC_URLS=SCRAPYD_SERVERS_PUBLIC_URLS or [''] * len(SCRAPYD_SERVERS),

            DAEMONSTATUS_REFRESH_INTERVAL=app.config.get('DAEMONSTATUS_REFRESH_INTERVAL', 10),
            ENABLE_AUTH=app.config.get('ENABLE_AUTH', False),
            SHOW_SCRAPYD_ITEMS=app.config.get('SHOW_SCRAPYD_ITEMS', True),
        )

    # To solve https://github.com/my8100/scrapydweb/issues/17
    # http://flask.pocoo.org/docs/1.0/cli/?highlight=flask_debug#environments
    # flask/helpers.py: get_env() The default is 'production'
    # On Windows, get/set/delete: set FLASK_ENV, set FLASK_ENV=production, set set FLASK_ENV=
    # if not os.environ.get('FLASK_ENV'):
        # os.environ['FLASK_ENV'] = 'development'
        # print("The environment variable 'FLASK_ENV' has been set to 'development'")
        # print("WARNING: Do not use the development server in a production. "
               # "Check out http://flask.pocoo.org/docs/1.0/deploying/")

    # http://flask.pocoo.org/docs/1.0/config/?highlight=flask_debug#environment-and-debug-features
    if app.config.get('DEBUG', False):
        os.environ['FLASK_DEBUG'] = '1'
        logger.info("Note that use_reloader is set to False in run.py")
    else:
        os.environ['FLASK_DEBUG'] = '0'

    # site-packages/flask/app.py
    # Threaded mode is enabled by default.
    # https://stackoverflow.com/a/28590266/10517783 to run in HTTP or HTTPS mode
    # site-packages/werkzeug/serving.py
    # https://stackoverflow.com/questions/13895176/sqlalchemy-and-sqlite-database-is-locked
    if app.config.get('ENABLE_HTTPS', False):
        protocol = 'https'
        context = (app.config['CERTIFICATE_FILEPATH'], app.config['PRIVATEKEY_FILEPATH'])
    else:
        protocol = 'http'
        context = None

    print("{star}Visit ScrapydWeb at {protocol}://127.0.0.1:{port} "
          "or {protocol}://IP-OF-THE-CURRENT-HOST:{port}{star}\n".format(
           star=STAR, protocol=protocol, port=app.config['SCRAPYDWEB_PORT']))
    logger.info("For running Flask in production, check out http://flask.pocoo.org/docs/1.0/deploying/")
    apscheduler_logger.setLevel(logging.DEBUG)
    app.run(host=app.config['SCRAPYDWEB_BIND'], port=app.config['SCRAPYDWEB_PORT'],
            ssl_context=context, use_reloader=False)


def load_custom_settings(config):
    path = find_scrapydweb_settings_py(SCRAPYDWEB_SETTINGS_PY, os.getcwd())

    if path:
        config['SCRAPYDWEB_SETTINGS_PY_PATH'] = path
        print(u"{star}Overriding custom settings from {path}{star}".format(star=STAR, path=handle_slash(path)))
        config.from_pyfile(path)
    else:
        logger.error("%s not found: ", SCRAPYDWEB_SETTINGS_PY)
        try:
            copyfile(config['DEFAULT_SETTINGS_PY_PATH'], config['SCRAPYDWEB_SETTINGS_PY_PATH'])
        except:
            sys.exit("\nPlease copy the 'default_settings.py' file from the path above to current working directory,\n"
                     "and rename it to '{file}'.\n"
                     "Then add your SCRAPYD_SERVERS in the config file and restart scrapydweb.\n".format(
                      file=SCRAPYDWEB_SETTINGS_PY))
        else:
            sys.exit("\nATTENTION:\nYou may encounter ERROR if there are any running timer tasks added in v1.2.0,\n"
                     "and you have to restart scrapydweb and manually edit the tasks to resume them.\n"
                     "\nThe config file '{file}' has been copied to current working directory.\n"
                     "Please add your SCRAPYD_SERVERS in the config file and restart scrapydweb.\n".format(
                      file=SCRAPYDWEB_SETTINGS_PY))


def parse_args(config):
    parser = argparse.ArgumentParser(description='ScrapydWeb -- %s' % __description__)

    SCRAPYDWEB_BIND = config.get('SCRAPYDWEB_BIND', '0.0.0.0')
    parser.add_argument(
        '-b', '--bind',
        default=SCRAPYDWEB_BIND,
        help=("current: %s, note that setting to 0.0.0.0 or IP-OF-THE-CURRENT-HOST would make ScrapydWeb server "
              "visible externally, otherwise, type '-b 127.0.0.1'") % SCRAPYDWEB_BIND
    )

    SCRAPYDWEB_PORT = config.get('SCRAPYDWEB_PORT', 5000)
    parser.add_argument(
        '-p', '--port',
        default=SCRAPYDWEB_PORT,
        help="current: %s, accept connections on the specified port" % SCRAPYDWEB_PORT
    )

    SCRAPYD_SERVERS = config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
    parser.add_argument(
        '-ss', '--scrapyd_server',
        action='append',
        help=("current: %s, type '-ss 127.0.0.1 -ss username:password@192.168.123.123:6801#group' "
              "to set up more than one Scrapyd server to manage. ") % SCRAPYD_SERVERS
    )

    ENABLE_AUTH = config.get('ENABLE_AUTH', False)
    parser.add_argument(
        '-da', '--disable_auth',
        action='store_true',
        help="current: ENABLE_AUTH = %s, append '--disable_auth' to disable basic auth for web UI" % ENABLE_AUTH
    )

    ENABLE_LOGPARSER = config.get('ENABLE_LOGPARSER', False)
    parser.add_argument(
        '-dlp', '--disable_logparser',
        action='store_true',
        help=("current: ENABLE_LOGPARSER = %s, append '--disable_logparser' to disable running LogParser "
              "as a subprocess at startup") % ENABLE_LOGPARSER
    )

    SCHEDULER_STATE = SCHEDULER_STATE_DICT[handle_metadata().get('scheduler_state', STATE_RUNNING)]
    parser.add_argument(
        '-sw', '--switch_scheduler_state',
        action='store_true',
        help=("current: %s, append '--switch_scheduler_state' to switch the state of scheduler "
              "for timer tasks") % SCHEDULER_STATE
    )

    ENABLE_MONITOR = config.get('ENABLE_MONITOR', False)
    parser.add_argument(
        '-dm', '--disable_monitor',
        action='store_true',
        help="current: ENABLE_MONITOR = %s, append '--disable_monitor' to disable monitor" % ENABLE_MONITOR
    )

    DEBUG = config.get('DEBUG', False)
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help=("current: DEBUG = %s, append '--debug' to enable debug mode "
              "and the debugger would be available in the browser") % DEBUG
    )

    VERBOSE = config.get('VERBOSE', False)
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help=("current: VERBOSE = %s, append '--verbose' to set the logging level to DEBUG "
              "for getting more information about how ScrapydWeb works") % VERBOSE
    )

    return parser.parse_args()


def update_app_config(config, args):
    logger.debug("Reading settings from command line: %s", args)

    config.update(dict(
        SCRAPYDWEB_BIND=args.bind,
        SCRAPYDWEB_PORT=args.port,
    ))

    # scrapyd_server would be None if the -ss argument is not passed in
    if args.scrapyd_server:
        config['SCRAPYD_SERVERS'] = args.scrapyd_server

    # action='store_true': default False
    if args.disable_auth:
        config['ENABLE_AUTH'] = False
    if args.disable_logparser:
        config['ENABLE_LOGPARSER'] = False
    if args.switch_scheduler_state:
        if handle_metadata().get('scheduler_state', STATE_RUNNING) == STATE_RUNNING:
            handle_metadata('scheduler_state', STATE_PAUSED)
        else:
            handle_metadata('scheduler_state', STATE_RUNNING)
    if args.disable_monitor:
        config['ENABLE_MONITOR'] = False
    if args.debug:
        config['DEBUG'] = True
    if args.verbose:
        config['VERBOSE'] = True


if __name__ == '__main__':
    main()
